from typing import List, Optional

from autogluon.common.utils.try_import import try_import_ray

from ..constants import (
    BEST_QUALITY,
    BINARY,
    DATA,
    DEFAULT,
    ENV,
    HIGH_QUALITY,
    MEDIUM_QUALITY,
    MODEL,
    MULTICLASS,
    OPTIM,
    REGRESSION,
)
from .registry import Registry

automm_presets = Registry("automm_presets")
matcher_presets = Registry("matcher_presets")


def get_default_hpo_setup():
    try_import_ray()
    from ray import tune

    default_hyperparameter_tune_kwargs = {
        "searcher": "bayes",
        "scheduler": "ASHA",
        "num_trials": 512,
    }

    default_tunable_hyperparameters = {
        "optim.lr": tune.loguniform(1e-5, 1e-2),
        "optim.optim_type": tune.choice(["adamw", "sgd"]),
        "optim.max_epochs": tune.choice(list(range(5, 31))),
        "env.batch_size": tune.choice([16, 32, 64, 128, 256]),
    }

    return default_tunable_hyperparameters, default_hyperparameter_tune_kwargs


def parse_presets_str(presets: str):
    use_hpo = False
    if presets.endswith("_hpo"):
        presets = presets[:-4]
        use_hpo = True

    return presets, use_hpo


@automm_presets.register()
def default(presets: str = DEFAULT):
    """
    Register the presets for problem types: binary, multiclass, classification, and regression.

    Parameters
    ----------
    presets
        The preset name.

    Returns
    -------
    hyperparameters
        The hyperparameters for a given preset.
    hyperparameter_tune_kwargs
        The hyperparameter tuning kwargs.
    """
    hyperparameters = {
        "model.names": [
            "ft_transformer",
            "timm_image",
            "hf_text",
            "document_transformer",
            "fusion_mlp",
        ],
    }
    hyperparameter_tune_kwargs = {}

    presets, use_hpo = parse_presets_str(presets)
    if use_hpo:
        default_tunable_hyperparameters, default_hyperparameter_tune_kwargs = get_default_hpo_setup()
        hyperparameters.update(default_tunable_hyperparameters)
        hyperparameter_tune_kwargs.update(default_hyperparameter_tune_kwargs)

    if presets in [HIGH_QUALITY, DEFAULT]:
        if use_hpo:
            from ray import tune

            hyperparameters.update(
                {
                    "env.per_gpu_batch_size": 2,  # Cover some corner cases of HPO on multimodal data.
                    "model.hf_text.checkpoint_name": tune.choice(
                        [
                            "google/electra-base-discriminator",
                            "google/flan-t5-base",
                            "microsoft/deberta-v3-small",
                            "roberta-base",
                            "albert-xlarge-v2",
                        ]
                    ),
                    "model.timm_image.checkpoint_name": tune.choice(
                        [
                            "swin_base_patch4_window7_224",
                            "convnext_base_in22ft1k",
                            "vit_base_patch16_clip_224.laion2b_ft_in12k_in1k",
                            "caformer_b36.sail_in22k_ft_in1k",
                        ]
                    ),
                    "model.document_transformer.checkpoint_name": "microsoft/layoutlmv3-base",
                }
            )
        else:
            hyperparameters.update(
                {
                    "model.hf_text.checkpoint_name": "google/electra-base-discriminator",
                    "model.timm_image.checkpoint_name": "caformer_b36.sail_in22k_ft_in1k",
                    "model.document_transformer.checkpoint_name": "microsoft/layoutlmv3-base",
                }
            )
    elif presets == MEDIUM_QUALITY:
        if use_hpo:
            from ray import tune

            hyperparameters.update(
                {
                    "model.hf_text.checkpoint_name": tune.choice(
                        [
                            "google/electra-small-discriminator",
                            "google/flan-t5-small",
                            "microsoft/deberta-v3-xsmall",
                            "albert-base-v2",
                            "microsoft/MiniLM-L12-H384-uncased",
                        ]
                    ),
                    "model.timm_image.checkpoint_name": tune.choice(
                        ["mobilenetv3_large_100", "gluon_resnet18_v1b", "maxvit_rmlp_pico_rw_256.sw_in1k"]
                    ),
                    "model.document_transformer.checkpoint_name": "microsoft/layoutlmv2-base-uncased",
                }
            )
        else:
            hyperparameters.update(
                {
                    "model.hf_text.checkpoint_name": "google/electra-small-discriminator",
                    "model.timm_image.checkpoint_name": "mobilenetv3_large_100",
                    "model.document_transformer.checkpoint_name": "microsoft/layoutlmv2-base-uncased",
                    "optim.lr": 4e-4,
                }
            )
    elif presets == BEST_QUALITY:
        hyperparameters.update({"env.per_gpu_batch_size": 1})
        if use_hpo:
            from ray import tune

            hyperparameters.update(
                {
                    "model.hf_text.checkpoint_name": tune.choice(
                        [
                            "microsoft/deberta-v3-base",
                            "google/flan-t5-large",
                            "google/electra-large-discriminator",
                            "roberta-large",
                        ]
                    ),
                    "model.timm_image.checkpoint_name": tune.choice(
                        [
                            "swin_large_patch4_window7_224",
                            "eva_large_patch14_336.in22k_ft_in22k_in1k",
                            "vit_large_patch14_clip_336.openai_ft_in12k_in1k",
                        ]
                    ),
                    "model.document_transformer.checkpoint_name": "microsoft/layoutlmv3-large",
                }
            )
        else:
            hyperparameters.update(
                {
                    "model.hf_text.checkpoint_name": "microsoft/deberta-v3-base",
                    "model.timm_image.checkpoint_name": "swin_large_patch4_window7_224",
                    "model.document_transformer.checkpoint_name": "microsoft/layoutlmv3-large",
                }
            )
    elif presets == "multilingual":
        hyperparameters.update(
            {
                "model.hf_text.checkpoint_name": "microsoft/mdeberta-v3-base",
                "optim.top_k": 1,
                "env.precision": "bf16-mixed",
                "env.per_gpu_batch_size": 4,
            }
        )
    else:
        raise ValueError(f"Unknown preset type: {presets}")

    return hyperparameters, hyperparameter_tune_kwargs


@automm_presets.register()
def few_shot_classification(presets: str = DEFAULT):
    """
    Register the presets for few_shot_classification.

    Parameters
    ----------
    presets
        The preset name.

    Returns
    -------
    hyperparameters
        The hyperparameters for a given preset.
    hyperparameter_tune_kwargs
        The hyperparameter tuning kwargs.
    """
    hyperparameters, hyperparameter_tune_kwargs = default(presets=presets)
    hyperparameters.update(
        {
            "model.hf_text.checkpoint_name": "sentence-transformers/all-mpnet-base-v2",
            "model.hf_text.pooling_mode": "mean",
            "model.names": ["hf_text", "clip"],
            "model.clip.checkpoint_name": "openai/clip-vit-large-patch14-336",
            "model.clip.image_size": 336,
            "env.inference_batch_size_ratio": 1,
        }
    )
    hyperparameter_tune_kwargs = {}

    return hyperparameters, hyperparameter_tune_kwargs


@automm_presets.register()
def zero_shot_image_classification(presets: str = DEFAULT):
    """
    Register the presets for zero_shot_image_classification.

    Parameters
    ----------
    presets
        The preset name.

    Returns
    -------
    hyperparameters
        The hyperparameters for a given preset.
    hyperparameter_tune_kwargs
        The hyperparameter tuning kwargs.
    """
    hyperparameters = {
        "model.names": ["clip"],
        "model.clip.max_text_len": 0,
    }
    hyperparameter_tune_kwargs = {}

    if presets in [DEFAULT, BEST_QUALITY]:
        hyperparameters.update(
            {
                "model.clip.checkpoint_name": "openai/clip-vit-large-patch14-336",
                "model.clip.image_size": 336,
                "env.inference_batch_size_ratio": 1,
            }
        )
    elif presets == HIGH_QUALITY:
        hyperparameters.update(
            {
                "model.clip.checkpoint_name": "openai/clip-vit-large-patch14",
                "env.inference_batch_size_ratio": 1,
            }
        )
    elif presets == MEDIUM_QUALITY:
        hyperparameters.update(
            {
                "model.clip.checkpoint_name": "openai/clip-vit-base-patch32",
            }
        )
    else:
        raise ValueError(f"Unknown preset type: {presets}")

    return hyperparameters, hyperparameter_tune_kwargs


@automm_presets.register()
def object_detection(presets: str = DEFAULT):
    """
    Register the presets for object_detection.

    Parameters
    ----------
    presets
        The preset name.

    Returns
    -------
    hyperparameters
        The hyperparameters for a given preset.
    hyperparameter_tune_kwargs
        The hyperparameter tuning kwargs.
    """
    hyperparameters = {
        "model.names": ["mmdet_image"],
        "model.mmdet_image.frozen_layers": [],
        "optim.patience": 20,
        "optim.val_check_interval": 1.0,
        "optim.check_val_every_n_epoch": 1,
        "env.batch_size": 32,
        "env.per_gpu_batch_size": 1,
        "env.num_workers": 2,
        "optim.lr": 1e-5,
        "optim.weight_decay": 1e-4,
        "optim.lr_mult": 10,
        "optim.lr_choice": "two_stages",
        "optim.lr_schedule": "multi_step",
        "optim.gradient_clip_val": 0.1,
        "optim.max_epochs": 60,
        "optim.warmup_steps": 0.0,
        "optim.top_k": 1,
        "optim.top_k_average_method": "best",
        "env.inference_batch_size_ratio": 1,
        "env.strategy": "ddp",
        "env.auto_select_gpus": True,  # Turn on for detection to return devices in a list, TODO: fix the extra GPU usage bug
        "env.num_gpus": -1,
        "optim.lr_decay": 0.9,
    }
    hyperparameter_tune_kwargs = {}

    presets, use_hpo = parse_presets_str(presets)
    if use_hpo:
        default_tunable_hyperparameters, default_hyperparameter_tune_kwargs = get_default_hpo_setup()
        hyperparameters.update(default_tunable_hyperparameters)
        hyperparameter_tune_kwargs.update(default_hyperparameter_tune_kwargs)

    if presets == MEDIUM_QUALITY:
        hyperparameters.update(
            {
                "model.mmdet_image.checkpoint_name": "yolox_l",
                "env.per_gpu_batch_size": 2,  # Works on 8G GPU
                "optim.lr": 5e-5,
                "optim.patience": 5,
                "optim.max_epochs": 50,
                "optim.val_check_interval": 1.0,
                "optim.check_val_every_n_epoch": 3,
                "optim.lr_mult": 100,
                "optim.weight_decay": 1e-3,
                "optim.lr_schedule": "cosine_decay",
                "optim.gradient_clip_val": 1,
            }
        )
    elif presets in [DEFAULT, HIGH_QUALITY]:
        hyperparameters.update(
            {
                "model.mmdet_image.checkpoint_name": "dino-4scale_r50_8xb2-12e_coco",
            }
        )
    elif presets == BEST_QUALITY:
        hyperparameters.update(
            {
                "model.mmdet_image.checkpoint_name": "dino-5scale_swin-l_8xb2-36e_coco",
            }
        )
    else:
        raise ValueError(f"Unknown preset type: {presets}")

    return hyperparameters, hyperparameter_tune_kwargs


@automm_presets.register()
def semantic_segmentation(presets: str = DEFAULT):
    """
    Register the presets for semantic_segmentation.

    Parameters
    ----------
    presets
        The preset name.

    Returns
    -------
    hyperparameters
        The hyperparameters for a given preset.
    hyperparameter_tune_kwargs
        The hyperparameter tuning kwargs.
    """
    hyperparameters = {
        "model.names": ["sam"],
        "model.sam.checkpoint_name": "facebook/sam-vit-huge",
        "env.batch_size": 4,
        "env.per_gpu_batch_size": 1,
        "env.inference_batch_size_ratio": 1,
        "env.strategy": "ddp_find_unused_parameters_true",
        "env.auto_select_gpus": False,
        "env.num_gpus": -1,
        "env.num_workers": 4,
        "env.precision": "16-mixed",
        "optim.lr": 1e-4,
        "optim.loss_func": "structure_loss",
        "optim.lr_decay": 0,
        "optim.lr_mult": 1,
        "optim.lr_choice": "single_stage",
        "optim.lr_schedule": "polynomial_decay",
        "optim.max_epochs": 30,
        "optim.top_k": 3,
        "optim.top_k_average_method": "best",
        "optim.warmup_steps": 0.0,
        "optim.weight_decay": 0.0001,
        "optim.patience": 10,
        "optim.val_check_interval": 1.0,
        "optim.check_val_every_n_epoch": 1,
        "optim.peft": "lora",
        "optim.lora.module_filter": [".*vision_encoder.*attn"],
        "optim.lora.filter": ["q", "v"],
        "optim.extra_trainable_params": [".*mask_decoder"],
        "optim.lora.r": 3,
        "optim.lora.alpha": 32,
    }
    hyperparameter_tune_kwargs = {}

    presets, use_hpo = parse_presets_str(presets)
    if use_hpo:
        default_tunable_hyperparameters, default_hyperparameter_tune_kwargs = get_default_hpo_setup()
        hyperparameters.update(default_tunable_hyperparameters)
        hyperparameter_tune_kwargs.update(default_hyperparameter_tune_kwargs)

    return hyperparameters, hyperparameter_tune_kwargs


@automm_presets.register()
def ocr_text_detection(presets: str = DEFAULT):
    """
    Register the presets for ocr_text_detection.

    Parameters
    ----------
    presets
        The preset name.

    Returns
    -------
    hyperparameters
        The hyperparameters for a given preset.
    hyperparameter_tune_kwargs
        The hyperparameter tuning kwargs.
    """
    hyperparameters = {
        "model.names": ["mmocr_text_detection"],
        "model.mmocr_text_detection.checkpoint_name": "TextSnake",
        "env.inference_batch_size_ratio": 1,
        "env.num_gpus": 1,
        "env.precision": 32,
    }
    hyperparameter_tune_kwargs = {}

    presets, use_hpo = parse_presets_str(presets)
    if use_hpo:
        default_tunable_hyperparameters, default_hyperparameter_tune_kwargs = get_default_hpo_setup()
        hyperparameters.update(default_tunable_hyperparameters)
        hyperparameter_tune_kwargs.update(default_hyperparameter_tune_kwargs)

    return hyperparameters, hyperparameter_tune_kwargs


@automm_presets.register()
def ocr_text_recognition(presets: str = DEFAULT):
    """
    Register the presets for ocr_text_recognition.

    Parameters
    ----------
    presets
        The preset name.

    Returns
    -------
    hyperparameters
        The hyperparameters for a given preset.
    hyperparameter_tune_kwargs
        The hyperparameter tuning kwargs.
    """
    hyperparameters = {
        "model.names": ["mmocr_text_recognition"],
        "model.mmocr_text_recognition.checkpoint_name": "ABINet",
        "env.inference_batch_size_ratio": 1,
        "env.num_gpus": 1,
        "env.precision": 32,
    }
    hyperparameter_tune_kwargs = {}
    presets, use_hpo = parse_presets_str(presets)
    if use_hpo:
        default_tunable_hyperparameters, default_hyperparameter_tune_kwargs = get_default_hpo_setup()
        hyperparameters.update(default_tunable_hyperparameters)
        hyperparameter_tune_kwargs.update(default_hyperparameter_tune_kwargs)

    return hyperparameters, hyperparameter_tune_kwargs


@automm_presets.register()
def feature_extraction(presets: str = DEFAULT):  # TODO: rename the problem type as text_feature_extraction?
    """
    Register the presets for feature_extraction.

    Parameters
    ----------
    presets
        The preset name.

    Returns
    -------
    hyperparameters
        The hyperparameters for a given preset.
    hyperparameter_tune_kwargs
        The hyperparameter tuning kwargs.
    """
    hyperparameters = {
        "model.names": ["hf_text"],
        "model.hf_text.checkpoint_name": "sentence-transformers/msmarco-MiniLM-L-12-v3",
        "model.hf_text.pooling_mode": "mean",
        "env.inference_batch_size_ratio": 1,
    }
    hyperparameter_tune_kwargs = {}

    return hyperparameters, hyperparameter_tune_kwargs


@automm_presets.register()
@matcher_presets.register()
def image_similarity(presets: str = DEFAULT):
    """
    Register the presets for image_similarity.

    Parameters
    ----------
    presets
        The preset name.

    Returns
    -------
    hyperparameters
        The hyperparameters for a given preset.
    hyperparameter_tune_kwargs
        The hyperparameter tuning kwargs.
    """
    hyperparameters = {
        "model.names": ["timm_image"],
    }
    hyperparameter_tune_kwargs = {}

    presets, use_hpo = parse_presets_str(presets)
    if use_hpo:
        default_tunable_hyperparameters, default_hyperparameter_tune_kwargs = get_default_hpo_setup()
        hyperparameters.update(default_tunable_hyperparameters)
        hyperparameter_tune_kwargs.update(default_hyperparameter_tune_kwargs)

    if presets in [DEFAULT, HIGH_QUALITY]:
        hyperparameters.update(
            {
                "model.timm_image.checkpoint_name": "caformer_b36.sail_in22k_ft_in1k",
            }
        )
    elif presets == MEDIUM_QUALITY:
        hyperparameters.update(
            {
                "model.timm_image.checkpoint_name": "swin_small_patch4_window7_224",
            }
        )
    elif presets == BEST_QUALITY:
        hyperparameters.update(
            {
                "model.timm_image.checkpoint_name": "swin_large_patch4_window7_224",
            }
        )
    else:
        raise ValueError(f"Unknown preset type: {presets}")

    return hyperparameters, hyperparameter_tune_kwargs


@automm_presets.register()
@matcher_presets.register()
def text_similarity(presets: str = DEFAULT):
    """
    Register the presets for text_similarity.

    Parameters
    ----------
    presets
        The preset name.

    Returns
    -------
    hyperparameters
        The hyperparameters for a given preset.
    hyperparameter_tune_kwargs
        The hyperparameter tuning kwargs.
    """
    hyperparameters = {
        "model.names": ["hf_text"],
        "model.hf_text.pooling_mode": "mean",
        "data.categorical.convert_to_text": True,
        "data.numerical.convert_to_text": True,
    }
    hyperparameter_tune_kwargs = {}

    presets, use_hpo = parse_presets_str(presets)
    if use_hpo:
        default_tunable_hyperparameters, default_hyperparameter_tune_kwargs = get_default_hpo_setup()
        hyperparameters.update(default_tunable_hyperparameters)
        hyperparameter_tune_kwargs.update(default_hyperparameter_tune_kwargs)

    if presets in [DEFAULT, HIGH_QUALITY]:
        hyperparameters.update(
            {
                "model.hf_text.checkpoint_name": "sentence-transformers/all-MiniLM-L12-v2",
            }
        )
    elif presets == MEDIUM_QUALITY:
        hyperparameters.update(
            {
                "model.hf_text.checkpoint_name": "sentence-transformers/all-MiniLM-L6-v2",
            }
        )
    elif presets == BEST_QUALITY:
        hyperparameters.update(
            {
                "model.hf_text.checkpoint_name": "sentence-transformers/all-mpnet-base-v2",
            }
        )
    else:
        raise ValueError(f"Unknown preset type: {presets}")

    return hyperparameters, hyperparameter_tune_kwargs


@automm_presets.register()
@matcher_presets.register()
def image_text_similarity(presets: str = DEFAULT):
    """
    Register the presets for image_text_similarity.

    Parameters
    ----------
    presets
        The preset name.

    Returns
    -------
    hyperparameters
        The hyperparameters for a given preset.
    hyperparameter_tune_kwargs
        The hyperparameter tuning kwargs.
    """
    hyperparameters = {
        "model.names": ["clip"],
        "matcher.loss.type": "multi_negatives_softmax_loss",
        "optim.lr": 1e-5,
    }
    hyperparameter_tune_kwargs = {}

    presets, use_hpo = parse_presets_str(presets)
    if use_hpo:
        default_tunable_hyperparameters, default_hyperparameter_tune_kwargs = get_default_hpo_setup()
        hyperparameters.update(default_tunable_hyperparameters)
        hyperparameter_tune_kwargs.update(default_hyperparameter_tune_kwargs)

    if presets in [DEFAULT, MEDIUM_QUALITY]:
        hyperparameters.update(
            {
                "model.clip.checkpoint_name": "openai/clip-vit-base-patch32",
                "env.per_gpu_batch_size": 128,
            }
        )
    elif presets == HIGH_QUALITY:
        hyperparameters.update(
            {
                "model.clip.checkpoint_name": "openai/clip-vit-large-patch14",
                "env.per_gpu_batch_size": 16,
            }
        )
    elif presets == BEST_QUALITY:
        hyperparameters.update(
            {
                "model.clip.checkpoint_name": "openai/clip-vit-large-patch14-336",
                "model.clip.image_size": 336,
                "env.per_gpu_batch_size": 8,
            }
        )
    else:
        raise ValueError(f"Unknown preset type: {presets}")

    return hyperparameters, hyperparameter_tune_kwargs


@automm_presets.register()
def ner(presets: str = DEFAULT):
    """
    Register the presets for ner.

    Parameters
    ----------
    presets
        The preset name.

    Returns
    -------
    hyperparameters
        The hyperparameters for a given preset.
    hyperparameter_tune_kwargs
        The hyperparameter tuning kwargs.
    """
    hyperparameters = {
        "model.names": [
            "ft_transformer",
            "timm_image",
            "ner_text",
            "fusion_ner",
        ],
    }
    hyperparameter_tune_kwargs = {}

    presets, use_hpo = parse_presets_str(presets)
    if use_hpo:
        default_tunable_hyperparameters, default_hyperparameter_tune_kwargs = get_default_hpo_setup()
        hyperparameters.update(default_tunable_hyperparameters)
        hyperparameter_tune_kwargs.update(default_hyperparameter_tune_kwargs)

    if presets in [DEFAULT, HIGH_QUALITY]:
        hyperparameters.update(
            {
                "model.ner_text.checkpoint_name": "microsoft/deberta-v3-base",
            }
        )
    elif presets == MEDIUM_QUALITY:
        hyperparameters.update(
            {
                "model.ner_text.checkpoint_name": "google/electra-small-discriminator",
            }
        )
    elif presets == BEST_QUALITY:
        hyperparameters.update(
            {
                "model.ner_text.checkpoint_name": "microsoft/deberta-v3-large",
                "env.per_gpu_batch_size": 4,
            }
        )
    else:
        raise ValueError(f"Unknown preset type: {presets}")

    return hyperparameters, hyperparameter_tune_kwargs


@automm_presets.register()
def ensemble(presets: str = DEFAULT):
    hyperparameters = {
        "lf_mlp": {
            "model.names": ["ft_transformer", "timm_image", "hf_text", "fusion_mlp"],
            "model.timm_image.train_transforms": ["resize_shorter_side", "center_crop"],
            "model.hf_text.text_trivial_aug_maxscale": 0,
            "data.categorical.convert_to_text": False,
            "data.numerical.convert_to_text": False,
            "optim.cross_modal_align": "null",
            "data.modality_dropout": 0,
            "model.timm_image.use_learnable_image": False,
            "optim.lemda.turn_on": False,
        },
        "lf_transformer": {
            "model.names": ["ft_transformer", "timm_image", "hf_text", "fusion_transformer"],
            "model.timm_image.train_transforms": ["resize_shorter_side", "center_crop"],
            "model.hf_text.text_trivial_aug_maxscale": 0,
            "data.categorical.convert_to_text": False,
            "data.numerical.convert_to_text": False,
            "optim.cross_modal_align": "null",
            "data.modality_dropout": 0,
            "model.timm_image.use_learnable_image": False,
            "optim.lemda.turn_on": False,
        },
        "lf_clip": {
            "model.names": ["ft_transformer", "clip_image", "clip_text", "fusion_mlp"],
            "model.clip_image.data_types": ["image"],
            "model.clip_text.data_types": ["text"],
            "model.clip_image.train_transforms": ["resize_shorter_side", "center_crop"],
            "model.clip_text.text_trivial_aug_maxscale": 0,
            "data.categorical.convert_to_text": False,
            "data.numerical.convert_to_text": False,
            "optim.cross_modal_align": "null",
            "data.modality_dropout": 0,
            "model.clip_image.use_learnable_image": False,
            "optim.lemda.turn_on": False,
        },
        "early_fusion": {
            "model.names": ["meta_transformer"],
            "model.meta_transformer.checkpoint_path": "null",
            "model.meta_transformer.train_transforms": ["resize_shorter_side", "center_crop"],
            "model.meta_transformer.text_trivial_aug_maxscale": 0,
            "data.categorical.convert_to_text": False,
            "data.numerical.convert_to_text": False,
            "optim.cross_modal_align": "null",
            "data.modality_dropout": 0,
            "model.meta_transformer.use_learnable_image": False,
            "optim.lemda.turn_on": False,
        },
        "convert_categorical_to_text": {
            "model.names": ["ft_transformer", "timm_image", "hf_text", "fusion_mlp"],
            "model.timm_image.train_transforms": ["resize_shorter_side", "center_crop"],
            "model.hf_text.text_trivial_aug_maxscale": 0,
            "data.categorical.convert_to_text": True,
            "data.categorical.convert_to_text_template": "latex",
            "data.numerical.convert_to_text": False,
            "optim.cross_modal_align": "null",
            "data.modality_dropout": 0,
            "model.timm_image.use_learnable_image": False,
            "optim.lemda.turn_on": False,
        },
        "convert_numeric_to_text": {
            "model.names": ["ft_transformer", "timm_image", "hf_text", "fusion_mlp"],
            "model.timm_image.train_transforms": ["resize_shorter_side", "center_crop"],
            "model.hf_text.text_trivial_aug_maxscale": 0,
            "data.categorical.convert_to_text": False,
            "data.numerical.convert_to_text": True,
            "optim.cross_modal_align": "null",
            "data.modality_dropout": 0,
            "model.timm_image.use_learnable_image": False,
            "optim.lemda.turn_on": False,
        },
        "cross_modal_align_pos_only": {
            "model.names": ["ft_transformer", "timm_image", "hf_text", "fusion_mlp"],
            "model.timm_image.train_transforms": ["resize_shorter_side", "center_crop"],
            "model.hf_text.text_trivial_aug_maxscale": 0,
            "data.categorical.convert_to_text": False,
            "data.numerical.convert_to_text": False,
            "optim.cross_modal_align": "positive_only",
            "optim.cross_modal_align_weight": 1,
            "data.modality_dropout": 0,
            "model.timm_image.use_learnable_image": False,
            "optim.lemda.turn_on": False,
        },
        "input_aug": {
            "model.names": ["ft_transformer", "timm_image", "hf_text", "fusion_mlp"],
            "model.timm_image.train_transforms": ["resize_shorter_side", "center_crop", "trivial_augment"],
            "model.hf_text.text_trivial_aug_maxscale": 0.1,
            "data.categorical.convert_to_text": False,
            "data.numerical.convert_to_text": False,
            "optim.cross_modal_align": "null",
            "data.modality_dropout": 0,
            "model.timm_image.use_learnable_image": False,
            "optim.lemda.turn_on": False,
        },
        "feature_aug_lemda": {
            "model.names": ["ft_transformer", "timm_image", "hf_text", "fusion_mlp"],
            "model.timm_image.train_transforms": ["resize_shorter_side", "center_crop"],
            "model.hf_text.text_trivial_aug_maxscale": 0,
            "data.categorical.convert_to_text": False,
            "data.numerical.convert_to_text": False,
            "optim.cross_modal_align": "null",
            "data.modality_dropout": 0,
            "model.timm_image.use_learnable_image": False,
            "optim.lemda.turn_on": True,
            "optim.automatic_optimization": False,
        },
        "modality_dropout": {
            "model.names": ["ft_transformer", "timm_image", "hf_text", "fusion_mlp"],
            "model.timm_image.train_transforms": ["resize_shorter_side", "center_crop"],
            "model.hf_text.text_trivial_aug_maxscale": 0,
            "data.categorical.convert_to_text": False,
            "data.numerical.convert_to_text": False,
            "optim.cross_modal_align": "null",
            "data.modality_dropout": 0.2,
            "model.timm_image.use_learnable_image": False,
            "optim.lemda.turn_on": False,
        },
        "learnable_image": {
            "model.names": ["ft_transformer", "timm_image", "hf_text", "fusion_mlp"],
            "model.timm_image.train_transforms": ["resize_shorter_side", "center_crop"],
            "model.hf_text.text_trivial_aug_maxscale": 0,
            "data.categorical.convert_to_text": False,
            "data.numerical.convert_to_text": False,
            "optim.cross_modal_align": "null",
            "data.modality_dropout": 0,
            "model.timm_image.use_learnable_image": True,
            "optim.lemda.turn_on": False,
        },
        "modality_dropout_and_learnable_image": {
            "model.names": ["ft_transformer", "timm_image", "hf_text", "fusion_mlp"],
            "model.timm_image.train_transforms": ["resize_shorter_side", "center_crop"],
            "model.hf_text.text_trivial_aug_maxscale": 0,
            "data.categorical.convert_to_text": False,
            "data.numerical.convert_to_text": False,
            "optim.cross_modal_align": "null",
            "data.modality_dropout": 0.2,
            "model.timm_image.use_learnable_image": True,
            "optim.lemda.turn_on": False,
        },
    }

    if presets in [DEFAULT, HIGH_QUALITY]:
        for v in hyperparameters.values():
            if "timm_image" in v["model.names"]:
                v["model.timm_image.checkpoint_name"] = "caformer_b36.sail_in22k_ft_in1k"
            if "hf_text" in v["model.names"]:
                v["model.hf_text.checkpoint_name"] = "google/electra-base-discriminator"
            if "meta_transformer" in v["model.names"]:
                v["model.meta_transformer.model_version"] = "base"
            if "clip_image" in v["model.names"]:
                v["model.clip_image.checkpoint_name"] = "openai/clip-vit-base-patch32"
            if "clip_text" in v["model.names"]:
                v["model.clip_text.checkpoint_name"] = "openai/clip-vit-base-patch32"

    elif presets == MEDIUM_QUALITY:
        for v in hyperparameters.values():
            if "timm_image" in v["model.names"]:
                v["model.timm_image.checkpoint_name"] = "mobilenetv3_large_100"
            if "hf_text" in v["model.names"]:
                v["model.hf_text.checkpoint_name"] = "google/electra-small-discriminator"
            if "meta_transformer" in v["model.names"]:
                v["model.meta_transformer.model_version"] = "base"
            if "clip_image" in v["model.names"]:
                v["model.clip_image.checkpoint_name"] = "openai/clip-vit-base-patch32"
            if "clip_text" in v["model.names"]:
                v["model.clip_text.checkpoint_name"] = "openai/clip-vit-base-patch32"
    elif presets == BEST_QUALITY:
        for v in hyperparameters.values():
            if "timm_image" in v["model.names"]:
                v["model.timm_image.checkpoint_name"] = "swin_large_patch4_window7_224"
            if "hf_text" in v["model.names"]:
                v["model.hf_text.checkpoint_name"] = "microsoft/deberta-v3-base"
            if "meta_transformer" in v["model.names"]:
                v["model.meta_transformer.model_version"] = "large"
            if "clip_image" in v["model.names"]:
                v["model.clip_image.checkpoint_name"] = "openai/clip-vit-large-patch14"
            if "clip_text" in v["model.names"]:
                v["model.clip_text.checkpoint_name"] = "openai/clip-vit-large-patch14"
    else:
        raise ValueError(f"Unknown preset type: {presets}")

    return hyperparameters, None


def list_presets(verbose: bool = False):
    """
    List all available presets.

    Returns
    -------
    A list of presets.
    """
    preset_keys = automm_presets.list_keys()
    if not verbose:
        return preset_keys

    preset_details = {}
    for k in preset_keys:
        preset_details[k] = automm_presets.create(k)

    return preset_details


def get_basic_config(extra: Optional[List[str]] = None):
    """
    Get the basic config of AutoMM.

    Parameters
    ----------
    extra
        A list of extra config keys.

    Returns
    -------
    A dict config with keys: MODEL, DATA, OPTIM, ENV, and their default values.
    """
    config = {
        MODEL: DEFAULT,
        DATA: DEFAULT,
        OPTIM: DEFAULT,
        ENV: DEFAULT,
    }
    if extra:
        for k in extra:
            config[k] = DEFAULT

    return config


def get_presets(problem_type: str, presets: str):
    """
    Get the default hyperparameters and hyperparameter_tune_kwargs given problem type and presets.

    Parameters
    ----------
    problem_type
        Problem type.
    presets
        Name of a preset.

    Returns
    -------
    hyperparameters
        The hyperparameter overrides of this preset.
    hyperparameter_tune_kwargs
        Hyperparameter tuning strategy and kwargs (for example, how many HPO trials to run).
    """
    if not presets:
        presets = DEFAULT
    if presets == "hpo":
        presets = f"{DEFAULT}_{presets}"
    presets = presets.lower()
    if problem_type in [
        BINARY,
        MULTICLASS,
        REGRESSION,
        None,
    ]:
        problem_type = DEFAULT

    if problem_type in automm_presets.list_keys():
        hyperparameters, hyperparameter_tune_kwargs = automm_presets.create(problem_type, presets)
    else:
        raise ValueError(
            f"Problem type '{problem_type}' doesn't have any presets yet. "
            f"Consider one of these: {automm_presets.list_keys()}"
        )

    return hyperparameters, hyperparameter_tune_kwargs


def get_ensemble_presets(presets):
    if not presets:
        presets = DEFAULT
    return automm_presets.create("ensemble", presets)



# -*- coding: utf-8 -*-

import numpy as np
import torch
import torchmetrics
from scipy.ndimage import convolve
from scipy.ndimage import distance_transform_edt as bwdist

_EPS = np.spacing(1)  # the different implementation of epsilon (extreme min value) between numpy and matlab
_TYPE = np.float64


def _prepare_data(pred: np.ndarray, gt: np.ndarray) -> tuple:
    """
    A numpy-based function for preparing ``pred`` and ``gt``.
    - for ``pred``, it looks like ``mapminmax(im2double(...))`` of matlab;
    - ``gt`` will be binarized by 128.
    :param pred: prediction
    :param gt: mask
    :return: pred, gt
    """
    gt = gt > 128
    # im2double, mapminmax
    pred = pred / 255
    if pred.max() != pred.min():
        pred = (pred - pred.min()) / (pred.max() - pred.min())
    return pred, gt


def _get_adaptive_threshold(matrix: np.ndarray, max_value: float = 1) -> float:
    """
    Return an adaptive threshold, which is equal to twice the mean of ``matrix``.
    :param matrix: a data array
    :param max_value: the upper limit of the threshold
    :return: min(2 * matrix.mean(), max_value)
    """
    return min(2 * matrix.mean(), max_value)


class Fmeasure(object):
    def __init__(self, beta: float = 1.0):
        """
        F-measure for SOD.
        ::
            @inproceedings{Fmeasure,
                title={Frequency-tuned salient region detection},
                author={Achanta, Radhakrishna and Hemami, Sheila and Estrada, Francisco and S{\"u}sstrunk, Sabine},
                booktitle=CVPR,
                number={CONF},
                pages={1597--1604},
                year={2009}
            }
        :param beta: the weight of the precision
        """
        self.beta = beta
        self.precisions = []
        self.recalls = []
        self.adaptive_fms = []
        self.changeable_fms = []

    def step(self, pred: np.ndarray, gt: np.ndarray):
        pred, gt = _prepare_data(pred, gt)

        adaptive_fm = self.cal_adaptive_fm(pred=pred, gt=gt)
        self.adaptive_fms.append(adaptive_fm)

        precisions, recalls, changeable_fms = self.cal_pr(pred=pred, gt=gt)
        self.precisions.append(precisions)
        self.recalls.append(recalls)
        self.changeable_fms.append(changeable_fms)

    def cal_adaptive_fm(self, pred: np.ndarray, gt: np.ndarray) -> float:
        """
        Calculate the adaptive F-measure.
        :return: adaptive_fm
        """
        # ``np.count_nonzero`` is faster and better
        adaptive_threshold = _get_adaptive_threshold(pred, max_value=1)
        binary_predcition = pred >= adaptive_threshold
        area_intersection = binary_predcition[gt].sum()
        if area_intersection == 0:
            adaptive_fm = 0
        else:
            pre = area_intersection / np.count_nonzero(binary_predcition)
            rec = area_intersection / np.count_nonzero(gt)
            adaptive_fm = (1 + self.beta) * pre * rec / (self.beta * pre + rec)
        return adaptive_fm

    def cal_pr(self, pred: np.ndarray, gt: np.ndarray) -> tuple:
        """
        Calculate the corresponding precision and recall when the threshold changes from 0 to 255.
        These precisions and recalls can be used to obtain the mean F-measure, maximum F-measure,
        precision-recall curve and F-measure-threshold curve.
        For convenience, ``changeable_fms`` is provided here, which can be used directly to obtain
        the mean F-measure, maximum F-measure and F-measure-threshold curve.
        :return: precisions, recalls, changeable_fms
        """
        pred = (pred * 255).astype(np.uint8)
        bins = np.linspace(0, 256, 257)
        fg_hist, _ = np.histogram(pred[gt], bins=bins)
        bg_hist, _ = np.histogram(pred[~gt], bins=bins)

        fg_w_thrs = np.cumsum(np.flip(fg_hist), axis=0)
        bg_w_thrs = np.cumsum(np.flip(bg_hist), axis=0)

        TPs = fg_w_thrs
        Ps = fg_w_thrs + bg_w_thrs

        Ps[Ps == 0] = 1
        T = max(np.count_nonzero(gt), 1)

        precisions = TPs / Ps
        recalls = TPs / T

        numerator = (1 + self.beta) * precisions * recalls
        denominator = np.where(numerator == 0, 1, self.beta * precisions + recalls)
        changeable_fms = numerator / denominator
        return precisions, recalls, changeable_fms

    def get_results(self) -> dict:
        """
        Return the results about F-measure.
        :return: dict(fm=dict(adp=adaptive_fm, curve=changeable_fm), pr=dict(p=precision, r=recall))
        """
        adaptive_fm = np.mean(np.array(self.adaptive_fms, _TYPE))
        changeable_fm = np.mean(np.array(self.changeable_fms, dtype=_TYPE), axis=0)
        precision = np.mean(np.array(self.precisions, dtype=_TYPE), axis=0)  # N, 256
        recall = np.mean(np.array(self.recalls, dtype=_TYPE), axis=0)  # N, 256
        return dict(fm=dict(adp=adaptive_fm, curve=changeable_fm), pr=dict(p=precision, r=recall))


class MAE_SOD(object):
    def __init__(self):
        """
        MAE(mean absolute error) for SOD.
        ::
            @inproceedings{MAE,
                title={Saliency filters: Contrast based filtering for salient region detection},
                author={Perazzi, Federico and Kr{\"a}henb{\"u}hl, Philipp and Pritch, Yael and Hornung, Alexander},
                booktitle=CVPR,
                pages={733--740},
                year={2012}
            }
        """
        self.maes = []

    def step(self, pred: np.ndarray, gt: np.ndarray):
        pred, gt = _prepare_data(pred, gt)

        mae = self.cal_mae(pred, gt)
        # mae = np.sum(cv2.absdiff(gt.astype(float), pred.astype(float))) / (pred.shape[1] * pred.shape[0])
        self.maes.append(mae)

    def cal_mae(self, pred: np.ndarray, gt: np.ndarray) -> np.ndarray:
        """
        Calculate the mean absolute error.
        :return: mae
        """
        mae = np.mean(np.abs(pred - gt))
        return mae

    def get_results(self) -> dict:
        """
        Return the results about MAE.
        :return: dict(mae=mae)
        """
        mae = np.mean(np.array(self.maes, _TYPE))
        return dict(mae=mae)


class Smeasure(object):
    def __init__(self, alpha: float = 0.5):
        """
        S-measure(Structure-measure) of SOD.
        ::
            @inproceedings{Smeasure,
                title={Structure-measure: A new way to eval foreground maps},
                author={Fan, Deng-Ping and Cheng, Ming-Ming and Liu, Yun and Li, Tao and Borji, Ali},
                booktitle=ICCV,
                pages={4548--4557},
                year={2017}
            }
        :param alpha: the weight for balancing the object score and the region score
        """
        self.sms = []
        self.alpha = alpha

    def step(self, pred: np.ndarray, gt: np.ndarray):
        pred, gt = _prepare_data(pred=pred, gt=gt)

        sm = self.cal_sm(pred, gt)
        self.sms.append(sm)

    def cal_sm(self, pred: np.ndarray, gt: np.ndarray) -> float:
        """
        Calculate the S-measure.
        :return: s-measure
        """
        y = np.mean(gt)
        if y == 0:
            sm = 1 - np.mean(pred)
        elif y == 1:
            sm = np.mean(pred)
        else:
            sm = self.alpha * self.object(pred, gt) + (1 - self.alpha) * self.region(pred, gt)
            sm = max(0, sm)
        return sm

    def object(self, pred: np.ndarray, gt: np.ndarray) -> float:
        """
        Calculate the object score.
        """
        fg = pred * gt
        bg = (1 - pred) * (1 - gt)
        u = np.mean(gt)
        object_score = u * self.s_object(fg, gt) + (1 - u) * self.s_object(bg, 1 - gt)
        return object_score

    def s_object(self, pred: np.ndarray, gt: np.ndarray) -> float:
        x = np.mean(pred[gt == 1])
        sigma_x = np.std(pred[gt == 1], ddof=1)
        score = 2 * x / (np.power(x, 2) + 1 + sigma_x + _EPS)
        return score

    def region(self, pred: np.ndarray, gt: np.ndarray) -> float:
        """
        Calculate the region score.
        """
        x, y = self.centroid(gt)
        part_info = self.divide_with_xy(pred, gt, x, y)
        w1, w2, w3, w4 = part_info["weight"]
        # assert np.isclose(w1 + w2 + w3 + w4, 1), (w1 + w2 + w3 + w4, pred.mean(), gt.mean())

        pred1, pred2, pred3, pred4 = part_info["pred"]
        gt1, gt2, gt3, gt4 = part_info["gt"]
        score1 = self.ssim(pred1, gt1)
        score2 = self.ssim(pred2, gt2)
        score3 = self.ssim(pred3, gt3)
        score4 = self.ssim(pred4, gt4)

        return w1 * score1 + w2 * score2 + w3 * score3 + w4 * score4

    def centroid(self, matrix: np.ndarray) -> tuple:
        """
        To ensure consistency with the matlab code, one is added to the centroid coordinate,
        so there is no need to use the redundant addition operation when dividing the region later,
        because the sequence generated by ``1:X`` in matlab will contain ``X``.
        :param matrix: a bool data array
        :return: the centroid coordinate
        """
        h, w = matrix.shape
        area_object = np.count_nonzero(matrix)
        if area_object == 0:
            x = np.round(w / 2)
            y = np.round(h / 2)
        else:
            # More details can be found at: https://www.yuque.com/lart/blog/gpbigm
            y, x = np.argwhere(matrix).mean(axis=0).round()
        return int(x) + 1, int(y) + 1

    def divide_with_xy(self, pred: np.ndarray, gt: np.ndarray, x: int, y: int) -> dict:
        """
        Use (x,y) to divide the ``pred`` and the ``gt`` into four submatrices, respectively.
        """
        h, w = gt.shape
        area = h * w

        gt_LT = gt[0:y, 0:x]
        gt_RT = gt[0:y, x:w]
        gt_LB = gt[y:h, 0:x]
        gt_RB = gt[y:h, x:w]

        pred_LT = pred[0:y, 0:x]
        pred_RT = pred[0:y, x:w]
        pred_LB = pred[y:h, 0:x]
        pred_RB = pred[y:h, x:w]

        w1 = x * y / area
        w2 = y * (w - x) / area
        w3 = (h - y) * x / area
        w4 = 1 - w1 - w2 - w3

        return dict(
            gt=(gt_LT, gt_RT, gt_LB, gt_RB),
            pred=(pred_LT, pred_RT, pred_LB, pred_RB),
            weight=(w1, w2, w3, w4),
        )

    def ssim(self, pred: np.ndarray, gt: np.ndarray) -> float:
        """
        Calculate the ssim score.
        """
        h, w = pred.shape
        N = h * w

        x = np.mean(pred)
        y = np.mean(gt)

        sigma_x = np.sum((pred - x) ** 2) / (N - 1)
        sigma_y = np.sum((gt - y) ** 2) / (N - 1)
        sigma_xy = np.sum((pred - x) * (gt - y)) / (N - 1)

        alpha = 4 * x * y * sigma_xy
        beta = (x**2 + y**2) * (sigma_x + sigma_y)

        if alpha != 0:
            score = alpha / (beta + _EPS)
        elif alpha == 0 and beta == 0:
            score = 1
        else:
            score = 0
        return score

    def get_results(self) -> dict:
        """
        Return the results about S-measure.
        :return: dict(sm=sm)
        """
        sm = np.mean(np.array(self.sms, dtype=_TYPE))
        return dict(sm=sm)


class Emeasure(object):
    def __init__(self):
        """
        E-measure(Enhanced-alignment Measure) for SOD.
        More details about the implementation can be found in https://www.yuque.com/lart/blog/lwgt38
        ::
            @inproceedings{Emeasure,
                title="Enhanced-alignment Measure for Binary Foreground Map Evaluation",
                author="Deng-Ping {Fan} and Cheng {Gong} and Yang {Cao} and Bo {Ren} and Ming-Ming {Cheng} and Ali {Borji}",
                booktitle=IJCAI,
                pages="698--704",
                year={2018}
            }
        """
        self.adaptive_ems = []
        self.changeable_ems = []

    def step(self, pred: np.ndarray, gt: np.ndarray):
        pred, gt = _prepare_data(pred=pred, gt=gt)

        self.gt_fg_numel = np.count_nonzero(gt)
        self.gt_size = gt.shape[0] * gt.shape[1]

        changeable_ems = self.cal_changeable_em(pred, gt)
        self.changeable_ems.append(changeable_ems)
        adaptive_em = self.cal_adaptive_em(pred, gt)
        self.adaptive_ems.append(adaptive_em)

    def cal_adaptive_em(self, pred: np.ndarray, gt: np.ndarray) -> float:
        """
        Calculate the adaptive E-measure.
        :return: adaptive_em
        """
        adaptive_threshold = _get_adaptive_threshold(pred, max_value=1)
        adaptive_em = self.cal_em_with_threshold(pred, gt, threshold=adaptive_threshold)
        return adaptive_em

    def cal_changeable_em(self, pred: np.ndarray, gt: np.ndarray) -> np.ndarray:
        """
        Calculate the changeable E-measure, which can be used to obtain the mean E-measure,
        the maximum E-measure and the E-measure-threshold curve.
        :return: changeable_ems
        """
        changeable_ems = self.cal_em_with_cumsumhistogram(pred, gt)
        return changeable_ems

    def cal_em_with_threshold(self, pred: np.ndarray, gt: np.ndarray, threshold: float) -> float:
        """
        Calculate the E-measure corresponding to the specific threshold.
        Variable naming rules within the function:
        ``[pred attribute(foreground fg, background bg)]_[gt attribute(foreground fg, background bg)]_[meaning]``
        If only ``pred`` or ``gt`` is considered, another corresponding attribute location is replaced with '``_``'.
        """
        binarized_pred = pred >= threshold
        fg_fg_numel = np.count_nonzero(binarized_pred & gt)
        fg_bg_numel = np.count_nonzero(binarized_pred & ~gt)

        fg___numel = fg_fg_numel + fg_bg_numel
        bg___numel = self.gt_size - fg___numel

        if self.gt_fg_numel == 0:
            enhanced_matrix_sum = bg___numel
        elif self.gt_fg_numel == self.gt_size:
            enhanced_matrix_sum = fg___numel
        else:
            parts_numel, combinations = self.generate_parts_numel_combinations(
                fg_fg_numel=fg_fg_numel,
                fg_bg_numel=fg_bg_numel,
                pred_fg_numel=fg___numel,
                pred_bg_numel=bg___numel,
            )

            results_parts = []
            for i, (part_numel, combination) in enumerate(zip(parts_numel, combinations)):
                align_matrix_value = (
                    2 * (combination[0] * combination[1]) / (combination[0] ** 2 + combination[1] ** 2 + _EPS)
                )
                enhanced_matrix_value = (align_matrix_value + 1) ** 2 / 4
                results_parts.append(enhanced_matrix_value * part_numel)
            enhanced_matrix_sum = sum(results_parts)

        em = enhanced_matrix_sum / (self.gt_size - 1 + _EPS)
        return em

    def cal_em_with_cumsumhistogram(self, pred: np.ndarray, gt: np.ndarray) -> np.ndarray:
        """
        Calculate the E-measure corresponding to the threshold that varies from 0 to 255..
        Variable naming rules within the function:
        ``[pred attribute(foreground fg, background bg)]_[gt attribute(foreground fg, background bg)]_[meaning]``
        If only ``pred`` or ``gt`` is considered, another corresponding attribute location is replaced with '``_``'.
        """
        pred = (pred * 255).astype(np.uint8)
        bins = np.linspace(0, 256, 257)
        fg_fg_hist, _ = np.histogram(pred[gt], bins=bins)
        fg_bg_hist, _ = np.histogram(pred[~gt], bins=bins)
        fg_fg_numel_w_thrs = np.cumsum(np.flip(fg_fg_hist), axis=0)
        fg_bg_numel_w_thrs = np.cumsum(np.flip(fg_bg_hist), axis=0)

        fg___numel_w_thrs = fg_fg_numel_w_thrs + fg_bg_numel_w_thrs
        bg___numel_w_thrs = self.gt_size - fg___numel_w_thrs

        if self.gt_fg_numel == 0:
            enhanced_matrix_sum = bg___numel_w_thrs
        elif self.gt_fg_numel == self.gt_size:
            enhanced_matrix_sum = fg___numel_w_thrs
        else:
            parts_numel_w_thrs, combinations = self.generate_parts_numel_combinations(
                fg_fg_numel=fg_fg_numel_w_thrs,
                fg_bg_numel=fg_bg_numel_w_thrs,
                pred_fg_numel=fg___numel_w_thrs,
                pred_bg_numel=bg___numel_w_thrs,
            )

            results_parts = np.empty(shape=(4, 256), dtype=np.float64)
            for i, (part_numel, combination) in enumerate(zip(parts_numel_w_thrs, combinations)):
                align_matrix_value = (
                    2 * (combination[0] * combination[1]) / (combination[0] ** 2 + combination[1] ** 2 + _EPS)
                )
                enhanced_matrix_value = (align_matrix_value + 1) ** 2 / 4
                results_parts[i] = enhanced_matrix_value * part_numel
            enhanced_matrix_sum = results_parts.sum(axis=0)

        em = enhanced_matrix_sum / (self.gt_size - 1 + _EPS)
        return em

    def generate_parts_numel_combinations(self, fg_fg_numel, fg_bg_numel, pred_fg_numel, pred_bg_numel):
        bg_fg_numel = self.gt_fg_numel - fg_fg_numel
        bg_bg_numel = pred_bg_numel - bg_fg_numel

        parts_numel = [fg_fg_numel, fg_bg_numel, bg_fg_numel, bg_bg_numel]

        mean_pred_value = pred_fg_numel / self.gt_size
        mean_gt_value = self.gt_fg_numel / self.gt_size

        demeaned_pred_fg_value = 1 - mean_pred_value
        demeaned_pred_bg_value = 0 - mean_pred_value
        demeaned_gt_fg_value = 1 - mean_gt_value
        demeaned_gt_bg_value = 0 - mean_gt_value

        combinations = [
            (demeaned_pred_fg_value, demeaned_gt_fg_value),
            (demeaned_pred_fg_value, demeaned_gt_bg_value),
            (demeaned_pred_bg_value, demeaned_gt_fg_value),
            (demeaned_pred_bg_value, demeaned_gt_bg_value),
        ]
        return parts_numel, combinations

    def get_results(self) -> dict:
        """
        Return the results about E-measure.
        :return: dict(em=dict(adp=adaptive_em, curve=changeable_em))
        """
        adaptive_em = np.mean(np.array(self.adaptive_ems, dtype=_TYPE))
        changeable_em = np.mean(np.array(self.changeable_ems, dtype=_TYPE), axis=0)
        return dict(em=dict(adp=adaptive_em, curve=changeable_em))


class WeightedFmeasure(object):
    def __init__(self, beta: float = 0.3):
        """
        Weighted F-measure for SOD.
        ::
            @inproceedings{wFmeasure,
                title={How to eval foreground maps?},
                author={Margolin, Ran and Zelnik-Manor, Lihi and Tal, Ayellet},
                booktitle=CVPR,
                pages={248--255},
                year={2014}
            }
        :param beta: the weight of the precision
        """
        self.beta = beta
        self.weighted_fms = []

    def step(self, pred: np.ndarray, gt: np.ndarray):
        pred, gt = _prepare_data(pred=pred, gt=gt)

        if np.all(~gt):
            wfm = 0
        else:
            wfm = self.cal_wfm(pred, gt)
        self.weighted_fms.append(wfm)

    def cal_wfm(self, pred: np.ndarray, gt: np.ndarray) -> float:
        """
        Calculate the weighted F-measure.
        """
        # [Dst,IDXT] = bwdist(dGT);
        Dst, Idxt = bwdist(gt == 0, return_indices=True)

        # %Pixel dependency
        # E = abs(FG-dGT);
        E = np.abs(pred - gt)
        # Et = E;
        # Et(~GT)=Et(IDXT(~GT)); %To deal correctly with the edges of the foreground region
        Et = np.copy(E)
        Et[gt == 0] = Et[Idxt[0][gt == 0], Idxt[1][gt == 0]]

        # K = fspecial('gaussian',7,5);
        # EA = imfilter(Et,K);
        K = self.matlab_style_gauss2D((7, 7), sigma=5)
        EA = convolve(Et, weights=K, mode="constant", cval=0)
        # MIN_E_EA = E;
        # MIN_E_EA(GT & EA<E) = EA(GT & EA<E);
        MIN_E_EA = np.where(gt & (EA < E), EA, E)

        # %Pixel importance
        # B = ones(size(GT));
        # B(~GT) = 2-1*exp(log(1-0.5)/5.*Dst(~GT));
        # Ew = MIN_E_EA.*B;
        B = np.where(gt == 0, 2 - np.exp(np.log(0.5) / 5 * Dst), np.ones_like(gt))
        Ew = MIN_E_EA * B

        # TPw = sum(dGT(:)) - sum(sum(Ew(GT)));
        # FPw = sum(sum(Ew(~GT)));
        TPw = np.sum(gt) - np.sum(Ew[gt == 1])
        FPw = np.sum(Ew[gt == 0])

        # R = 1- mean2(Ew(GT)); %Weighed Recall
        # P = TPw./(eps+TPw+FPw); %Weighted Precision
        # 注意这里使用mask索引矩阵的时候不可使用Ew[gt]，这实际上仅在索引Ew的0维度
        R = 1 - np.mean(Ew[gt == 1])
        P = TPw / (TPw + FPw + _EPS)

        # % Q = (1+Beta^2)*(R*P)./(eps+R+(Beta.*P));
        Q = (1 + self.beta) * R * P / (R + self.beta * P + _EPS)

        return Q

    def matlab_style_gauss2D(self, shape: tuple = (7, 7), sigma: int = 5) -> np.ndarray:
        """
        2D gaussian mask - should give the same result as MATLAB's
        fspecial('gaussian',[shape],[sigma])
        """
        m, n = [(ss - 1) / 2 for ss in shape]
        y, x = np.ogrid[-m : m + 1, -n : n + 1]
        h = np.exp(-(x * x + y * y) / (2 * sigma * sigma))
        h[h < np.finfo(h.dtype).eps * h.max()] = 0
        sumh = h.sum()
        if sumh != 0:
            h /= sumh
        return h

    def get_results(self) -> dict:
        """
        Return the results about weighted F-measure.
        :return: dict(wfm=weighted_fm)
        """
        weighted_fm = np.mean(np.array(self.weighted_fms, dtype=_TYPE))
        return dict(wfm=weighted_fm)


class Multiclass_IoU(torchmetrics.Metric):
    """
    Compute the IoU for multi-class semantic segmentation based on https://github.com/xieenze/Trans2Seg/blob/master/segmentron/utils/score.py.
    The direct use of torchmetrics for large dataset will lead to issues such as high CPU usage or insufficient memory.
    """

    def __init__(self, num_classes):
        super().__init__()
        self.add_state("total_inter", default=torch.zeros(num_classes), dist_reduce_fx=None)
        self.add_state("total_union", default=torch.zeros(num_classes), dist_reduce_fx=None)
        self.num_classes = num_classes

    def update(self, logits, labels):
        inter, union = self.batch_intersection_union(logits, labels)
        self.total_inter += inter
        self.total_union += union

    def compute(self):
        IoU = 1.0 * self.total_inter / (2.220446049250313e-16 + self.total_union)
        return torch.tensor(IoU.mean().item())

    def batch_intersection_union(self, output, target):
        mini = 1
        maxi = self.num_classes
        nbins = self.num_classes
        predict = torch.argmax(output, 1) + 1
        target = target.float() + 1

        predict = predict.float() * (target > 0).float()
        intersection = predict * (predict == target).float()
        # areas of intersection and union
        area_inter = torch.histc(intersection, bins=nbins, min=mini, max=maxi)
        area_pred = torch.histc(predict, bins=nbins, min=mini, max=maxi)
        area_lab = torch.histc(target, bins=nbins, min=mini, max=maxi)
        area_union = area_pred + area_lab - area_inter
        assert torch.sum(area_inter > area_union).item() == 0, "Intersection area should be smaller than Union area"
        return area_inter.float(), area_union.float()


class Binary_IoU(torchmetrics.Metric):
    """
    Compute the IoU for binary semantic segmentation. The direct use of torchmetrics to calculate IoU for multiple samples does not yield accurate results.
    So we iteratively calculate metric values and then take the average.
    """

    def __init__(
        self,
    ):
        super().__init__()
        self.add_state("logits", default=[], dist_reduce_fx=None)
        self.add_state("labels", default=[], dist_reduce_fx=None)

    def update(self, logits, labels):
        self.logits.append(logits)
        self.labels.append(labels)

    def compute(self):
        logits = torch.cat(self.logits).cpu()
        labels = torch.cat(self.labels).cpu()

        res_list = []
        metric = torchmetrics.JaccardIndex(task="binary")
        for logit, label in zip(logits, labels):
            res_list.append(metric(logit, label))
        return torch.mean(torch.tensor(res_list))


class Balanced_Error_Rate(torchmetrics.Metric):
    """
    Compute the balanced error rate.
    """

    def __init__(
        self,
    ):
        super().__init__()
        self.add_state("logits", default=[], dist_reduce_fx=None)
        self.add_state("labels", default=[], dist_reduce_fx=None)

    def update(self, logits, labels):
        self.logits.append(logits)
        self.labels.append(labels)

    def compute(self):
        logits = torch.cat(self.logits).cpu()
        labels = torch.cat(self.labels).cpu()

        labels = (labels * 255) > 125
        logits = (logits * 255) > 125
        ber = 1 - torchmetrics.Accuracy(
            task="multiclass", num_classes=2, average="macro", multidim_average="samplewise"
        )(logits, labels)
        return torch.mean(ber)


class COD(torchmetrics.Metric):
    def __init__(self):
        super().__init__()
        self.add_state("logits", default=[], dist_reduce_fx=None)
        self.add_state("labels", default=[], dist_reduce_fx=None)

    def update(self, logits, labels):
        self.logits.append(logits)
        self.labels.append(labels)

    def compute(self):
        pass


class SM(COD):
    def compute(self):
        logits = torch.cat(self.logits)
        labels = torch.cat(self.labels)
        assert logits.shape == labels.shape
        batchsize = labels.shape[0]

        metric_SM = Smeasure()

        for i in range(batchsize):
            true, pred = labels[i, 0].cpu().data.numpy() * 255, logits[i, 0].cpu().data.numpy() * 255
            metric_SM.step(pred=pred, gt=true)

        return torch.tensor(metric_SM.get_results()["sm"])


class FM(COD):
    def compute(self):
        logits = torch.cat(self.logits)
        labels = torch.cat(self.labels)
        assert logits.shape == labels.shape
        batchsize = labels.shape[0]

        metric_WFM = WeightedFmeasure()
        for i in range(batchsize):
            true, pred = labels[i, 0].cpu().data.numpy() * 255, logits[i, 0].cpu().data.numpy() * 255

            metric_WFM.step(pred=pred, gt=true)

        return torch.tensor(metric_WFM.get_results()["wfm"])


class EM(COD):
    def compute(self):
        logits = torch.cat(self.logits)
        labels = torch.cat(self.labels)
        assert logits.shape == labels.shape
        batchsize = labels.shape[0]

        metric_EM = Emeasure()

        for i in range(batchsize):
            true, pred = labels[i, 0].cpu().data.numpy() * 255, logits[i, 0].cpu().data.numpy() * 255

            metric_EM.step(pred=pred, gt=true)

        return torch.tensor(metric_EM.get_results()["em"]["curve"].mean())


class MAE(COD):
    def compute(self):
        logits = torch.cat(self.logits)
        labels = torch.cat(self.labels)
        assert logits.shape == labels.shape
        batchsize = labels.shape[0]

        metric_MAE = MAE_SOD()
        for i in range(batchsize):
            true, pred = labels[i, 0].cpu().data.numpy() * 255, logits[i, 0].cpu().data.numpy() * 255

            metric_MAE.step(pred=pred, gt=true)

        return torch.tensor(metric_MAE.get_results()["mae"])


COD_METRICS_NAMES = {"sm": SM(), "fm": FM(), "em": EM(), "mae": MAE()}


# TODO: Modify multi-gpu evaluation error. Maybe there will be a more elegant way.
class Multiclass_IoU_Pred:
    """
    Compute the IoU for multi-class semantic segmentation based on https://github.com/xieenze/Trans2Seg/blob/master/segmentron/utils/score.py.
    The direct use of torchmetrics for large dataset will lead to issues such as high CPU usage or insufficient memory.
    """

    def __init__(self, num_classes):
        super().__init__()
        self.total_inter = torch.zeros(num_classes)
        self.total_union = torch.zeros(num_classes)
        self.num_classes = num_classes

    def update(self, logits, labels):
        inter, union = self.batch_intersection_union(logits, labels)
        self.total_inter += inter
        self.total_union += union

    def compute(self):
        IoU = 1.0 * self.total_inter / (2.220446049250313e-16 + self.total_union)
        return torch.tensor(IoU.mean().item())

    def batch_intersection_union(self, output, target):
        mini = 1
        maxi = self.num_classes
        nbins = self.num_classes
        predict = torch.argmax(output, 1) + 1
        target = target.float() + 1

        predict = predict.float() * (target > 0).float()
        intersection = predict * (predict == target).float()
        # areas of intersection and union
        area_inter = torch.histc(intersection, bins=nbins, min=mini, max=maxi)
        area_pred = torch.histc(predict, bins=nbins, min=mini, max=maxi)
        area_lab = torch.histc(target, bins=nbins, min=mini, max=maxi)
        area_union = area_pred + area_lab - area_inter
        assert torch.sum(area_inter > area_union).item() == 0, "Intersection area should be smaller than Union area"
        return area_inter.float(), area_union.float()


class Binary_IoU_Pred:
    """
    Compute the IoU for binary semantic segmentation. The direct use of torchmetrics to calculate IoU for multiple samples does not yield accurate results.
    So we iteratively calculate metric values and then take the average.
    """

    def __init__(
        self,
    ):
        super().__init__()
        self.logits = []
        self.labels = []

    def update(self, logits, labels):
        self.logits.append(logits)
        self.labels.append(labels)

    def compute(self):
        logits = torch.cat(self.logits).cpu()
        labels = torch.cat(self.labels).cpu()

        res_list = []
        metric = torchmetrics.JaccardIndex(task="binary")
        for logit, label in zip(logits, labels):
            res_list.append(metric(logit, label))
        return torch.mean(torch.tensor(res_list))


class Balanced_Error_Rate_Pred:
    """
    Compute the balanced error rate.
    """

    def __init__(
        self,
    ):
        super().__init__()
        self.logits = []
        self.labels = []

    def update(self, logits, labels):
        self.logits.append(logits)
        self.labels.append(labels)

    def compute(self):
        logits = torch.cat(self.logits).cpu()
        labels = torch.cat(self.labels).cpu()

        labels = (labels * 255) > 125
        logits = (logits * 255) > 125
        ber = 1 - torchmetrics.Accuracy(
            task="multiclass", num_classes=2, average="macro", multidim_average="samplewise"
        )(logits, labels)
        return torch.mean(ber)


class COD_Pred:
    def __init__(self):
        super().__init__()
        self.logits = []
        self.labels = []

    def update(self, logits, labels):
        self.logits.append(logits)
        self.labels.append(labels)

    def compute(self):
        pass

    def reset(self):
        self.logits = []
        self.labels = []


class SM_Pred(COD_Pred):
    def compute(self):
        logits = torch.cat(self.logits)
        labels = torch.cat(self.labels)
        assert logits.shape == labels.shape

        batchsize = labels.shape[0]

        metric_SM = Smeasure()

        for i in range(batchsize):
            true, pred = labels[i, 0].cpu().data.numpy() * 255, logits[i, 0].cpu().data.numpy() * 255
            metric_SM.step(pred=pred, gt=true)
        self.reset()
        return torch.tensor(metric_SM.get_results()["sm"])


class FM_Pred(COD_Pred):
    def compute(self):
        logits = torch.cat(self.logits)
        labels = torch.cat(self.labels)
        assert logits.shape == labels.shape
        batchsize = labels.shape[0]

        metric_WFM = WeightedFmeasure()
        for i in range(batchsize):
            true, pred = labels[i, 0].cpu().data.numpy() * 255, logits[i, 0].cpu().data.numpy() * 255

            metric_WFM.step(pred=pred, gt=true)
        self.reset()
        return torch.tensor(metric_WFM.get_results()["wfm"])


class EM_Pred(COD_Pred):
    def compute(self):
        logits = torch.cat(self.logits)
        labels = torch.cat(self.labels)
        assert logits.shape == labels.shape
        batchsize = labels.shape[0]

        metric_EM = Emeasure()

        for i in range(batchsize):
            true, pred = labels[i, 0].cpu().data.numpy() * 255, logits[i, 0].cpu().data.numpy() * 255

            metric_EM.step(pred=pred, gt=true)
        self.reset()
        return torch.tensor(metric_EM.get_results()["em"]["curve"].mean())


class MAE_Pred(COD_Pred):
    def compute(self):
        logits = torch.cat(self.logits)
        labels = torch.cat(self.labels)
        assert logits.shape == labels.shape
        batchsize = labels.shape[0]

        metric_MAE = MAE_SOD()
        for i in range(batchsize):
            true, pred = labels[i, 0].cpu().data.numpy() * 255, logits[i, 0].cpu().data.numpy() * 255

            metric_MAE.step(pred=pred, gt=true)
        self.reset()
        return torch.tensor(metric_MAE.get_results()["mae"])


COD_METRICS_NAMES_Pred = {"sm": SM_Pred(), "fm": FM_Pred(), "em": EM_Pred(), "mae": MAE_Pred()}



from typing import Any, Dict, List, Optional, Tuple, Union

import torch
import torchmetrics


class CustomHitRate(torchmetrics.Metric):
    """
    Compute the hit rate when doing semantic search between two group of embeddings.
    We assume that (a_i, p_i) are a positive pair and (a_i, p_j) for i!=j a negative pair.
    """

    def __init__(
        self,
    ):
        super().__init__()
        self.add_state("query_embeddings", default=[], dist_reduce_fx=None)
        self.add_state("response_embeddings", default=[], dist_reduce_fx=None)
        self.add_state("logit_scale", default=[], dist_reduce_fx=None)

    def update(
        self,
        batch_query_embeds: torch.Tensor,
        batch_response_embeds: torch.Tensor,
        logit_scale: Optional[torch.Tensor] = None,
    ):
        self.query_embeddings.append(batch_query_embeds)
        self.response_embeddings.append(batch_response_embeds)
        if logit_scale is not None:
            self.logit_scale.append(logit_scale)

    def compute(self):
        query_embeddings = torch.cat(self.query_embeddings)
        response_embeddings = torch.cat(self.response_embeddings)
        if self.logit_scale:
            logit_scale = torch.mean(torch.stack(self.logit_scale))
        else:
            logit_scale = 1

        return self.compute_hit_rate(query_embeddings, response_embeddings, logit_scale)

    @staticmethod
    def compute_hit_rate(features_a, features_b, logit_scale, top_ks=[1, 5, 10]):
        """
        Compute symmetric hit rates between two groups of features.

        Parameters
        ----------
        features_a
            One group of features.
        features_b
            The other group of features.
        logit_scale
            The scale of logit (Used in CLIP).
        top_ks
            Consider only the top k elements for each query.

        Returns
        -------
        The accumulated hit rate.
        """
        assert len(features_a) == len(features_b)
        hit_rate = 0
        logits_per_a = (logit_scale * features_a @ features_b.t()).detach().cpu()
        logits_per_b = logits_per_a.t().detach().cpu()

        logits = {"logits_per_a": logits_per_a, "logits_per_b": logits_per_b}
        ground_truth = torch.arange(len(features_b)).view(-1, 1)

        for name, logit in logits.items():
            ranking = torch.argsort(logit, descending=True)
            preds = torch.where(ranking == ground_truth)[1]

            for k in top_ks:
                hit_rate += (preds < k).float().mean()

        hit_rate /= len(top_ks) * len(logits)
        return hit_rate



import logging
import os
import pickle
import zipfile

from ..constants import LAST_CHECKPOINT, MODEL_CHECKPOINT

logger = logging.getLogger(__name__)


def get_dir_ckpt_paths(path: str):
    """
    Get the dir path and ckpt path from a path.

    Parameters
    ----------
    path
        A path which can be either a dir or ckpt path.

    Returns
    -------
    The dir and ckpt paths.
    """
    path = os.path.abspath(os.path.expanduser(path))
    if os.path.isfile(path):
        dir_path = os.path.dirname(path)
        ckpt_path = path
    else:
        dir_path = path
        ckpt_path = None

    return dir_path, ckpt_path


def get_load_ckpt_paths(ckpt_path: str, dir_path: str, resume: bool):
    """
    Get the load_path and ckpt_path. They can be the same or different.
    #TODO: merging load_path and ckpt_path.

    Parameters
    ----------
    ckpt_path
        The path of one checkpoint, which can be None.
    dir_path
        The dir path from where to load model.
    resume
        Whether to resume training.

    Returns
    -------
    load_path and ckpt_path
    """
    if ckpt_path:
        load_path = ckpt_path
        logger.info(f"Loading checkpoint: '{ckpt_path}'")
    else:
        resume_ckpt_path = os.path.join(dir_path, LAST_CHECKPOINT)
        final_ckpt_path = os.path.join(dir_path, MODEL_CHECKPOINT)
        if resume:  # resume training which crashed before
            if not os.path.isfile(resume_ckpt_path):
                if os.path.isfile(final_ckpt_path):
                    raise ValueError(
                        f"Resuming checkpoint '{resume_ckpt_path}' doesn't exist, but "
                        f"final checkpoint '{final_ckpt_path}' exists, which means training "
                        f"is already completed."
                    )
                else:
                    raise ValueError(
                        f"Resuming checkpoint '{resume_ckpt_path}' and "
                        f"final checkpoint '{final_ckpt_path}' both don't exist. "
                        f"Consider starting training from scratch."
                    )
            load_path = resume_ckpt_path
            logger.info(f"Resume training from checkpoint: '{resume_ckpt_path}'")
            ckpt_path = resume_ckpt_path
        else:  # load a model checkpoint for prediction, evaluation, or continuing training on new data
            if not os.path.isfile(final_ckpt_path):
                if os.path.isfile(resume_ckpt_path):
                    raise ValueError(
                        f"Final checkpoint '{final_ckpt_path}' doesn't exist, but "
                        f"resuming checkpoint '{resume_ckpt_path}' exists, which means training "
                        f"is not done yet. Consider resume training from '{resume_ckpt_path}'."
                    )
                else:
                    raise ValueError(
                        f"Resuming checkpoint '{resume_ckpt_path}' and "
                        f"final checkpoint '{final_ckpt_path}' both don't exist. "
                        f"Consider starting training from scratch."
                    )
            load_path = final_ckpt_path
            logger.info(f"Load pretrained checkpoint: {os.path.join(dir_path, MODEL_CHECKPOINT)}")
            ckpt_path = None  # must set None since we do not resume training

    return load_path, ckpt_path


class CustomUnpickler(pickle.Unpickler):
    """
    This is to make pickle loading an object backward compatible.
    A df_preprocessor object saved with old name space `xxx.yyy` has errors
    when being loaded under the context of new name `aaa.bbb`.
    """

    def find_class(self, module, name):
        renamed_module = module
        if module.startswith("autogluon.text.automm"):
            renamed_module = module.replace("autogluon.text.automm", "autogluon.multimodal")

        return super(CustomUnpickler, self).find_class(renamed_module, name)


def protected_zip_extraction(zipfile_path, sha1_hash, folder):
    """
    Extract zip file to the folder.

    A signature file named ".SHA1HASH.sig" will be created if the extraction has been finished.

    Returns
    -------
    folder
        The directory to extract the zipfile
    """
    os.makedirs(folder, exist_ok=True)

    if sha1_hash:
        sha1_hash = sha1_hash[:6]
        signature = ".{}.sig".format(sha1_hash)

        if os.path.exists(os.path.join(folder, signature)):
            # We have found the signature file. Thus, we will not extract again.
            return folder
    else:
        signature = None

    # Extract the file
    logging.info("Extract files...")
    with zipfile.ZipFile(zipfile_path, "r") as zip_ref:
        zip_ref.extractall(folder)

    if signature:
        # Create the signature
        with open(os.path.join(folder, signature), "w"):
            pass

    return folder



import logging
import os
from datetime import datetime
from typing import Dict, List, Optional, Tuple, Union

import pytz

from ..constants import LAST_CHECKPOINT, MODEL_CHECKPOINT

logger = logging.getLogger(__name__)



import logging
import os
import time
import warnings
from typing import Optional

import torch
from torch import nn

from ..constants import BBOX, BBOX_FORMATS, COLUMN, IMAGE, IMAGE_VALID_NUM, LABEL, XYXY
from .utils import freeze_model_layers, lookup_mmdet_config, update_mmdet_config

try:
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        import mmcv
    import mmdet
    import mmengine
    from mmdet.registry import MODELS
    from mmengine.runner import load_checkpoint
except ImportError as e:
    mmcv = None
    mmdet = None
    mmengine = None


logger = logging.getLogger(__name__)


class MMDetAutoModelForObjectDetection(nn.Module):
    """
    Support MMDET object detection models.
    Refer to https://github.com/open-mmlab/mmdetection
    """

    def __init__(
        self,
        prefix: str,
        checkpoint_name: str,
        config_file: Optional[str] = None,
        classes: Optional[list] = None,
        pretrained: Optional[bool] = True,
        output_bbox_format: Optional[str] = XYXY,
        frozen_layers: Optional[list] = None,
    ):
        """
        Load a pretrained object detector from MMdetection.

        Parameters
        ----------
        prefix
            The prefix of the MMdetAutoModelForObjectDetection model.
        checkpoint_name
            Name of the mmdet checkpoint.
        classes
            All classes in this dataset.
        pretrained
            Whether using the pretrained mmdet models. If pretrained=True, download the pretrained model.
        """
        from ..utils import check_if_packages_installed

        check_if_packages_installed(package_names=["mmcv", "mmengine", "mmdet"])

        super().__init__()
        self.prefix = prefix
        self.pretrained = pretrained
        self.checkpoint = None
        self.checkpoint_name = checkpoint_name
        self.config_file = config_file
        self.classes = classes
        # Based on our offline benchmarking results, instead of freezing layers,
        # Setting backbone to a smaller learning rate achieves better results,
        self.frozen_layers = frozen_layers
        self._assign_backbone_layers()

        self.device = None

        if output_bbox_format.lower() in BBOX_FORMATS:
            self.output_bbox_format = output_bbox_format.lower()
        else:
            raise ValueError(
                f"Not supported bounding box output format for object detection: {output_bbox_format}. All supported bounding box output formats are: {BBOX_FORMATS}."
            )

        # TODO: Config only init (without checkpoint)

        self._get_checkpoint_and_config_file(checkpoint_name=checkpoint_name, config_file=config_file)
        self._load_config()

        self._update_classes(classes)
        self._load_checkpoint(self.checkpoint_file)

        freeze_model_layers(self.model, self.frozen_layers)

    def _reset_classes(self, classes: list):
        temp_ckpt_file = f"temp_ckpt_{int(time.time()*1000)}.pth"
        self._save_weights(temp_ckpt_file)
        self._update_classes(classes)
        self._load_checkpoint()
        os.remove(temp_ckpt_file)

    def _update_classes(self, classes: Optional[list] = None):
        if classes:
            self.num_classes = len(classes)
            self.classes = classes
            update_mmdet_config(key="num_classes", value=self.num_classes, config=self.config)
        else:
            self.num_classes = lookup_mmdet_config(key="num_classes", config=self.config)
            if not self.num_classes:
                self.num_classes = 1
                warnings.warn(
                    f"num_classes is not provided and is set to default value {self.num_classes} and this may cause error. Please provide sample_data_path in predictor's initialization."
                )
            self.classes = None
        self.id2label = dict(zip(range(self.num_classes), range(self.num_classes)))

    def _load_checkpoint(self, checkpoint_file):
        # build model and load pretrained weights
        from mmdet.utils import register_all_modules

        register_all_modules()  # https://github.com/open-mmlab/mmdetection/issues/9719

        self.model = MODELS.build(self.config.model)
        # yolox use self.config.model.data_preprocessor, yolov3 use self.config.data_preprocessor
        self.data_preprocessor = MODELS.build(
            self.config.data_preprocessor
            if "data_preprocessor" in self.config
            else self.config.model.data_preprocessor
        )

        if self.pretrained and checkpoint_file is not None:  # TODO: enable training from scratch
            self.checkpoint = load_checkpoint(self.model, checkpoint_file, map_location="cpu")

        # save the config and classes in the model for convenience
        self.model.cfg = self.config
        if self.classes:
            self.model.CLASSES = self.classes
        else:
            if self.checkpoint and "CLASSES" in self.checkpoint.get("meta", {}):
                warnings.simplefilter("once")
                warnings.warn(
                    f"Using classes provided in checkpoints: {self.checkpoint['meta']['CLASSES']}. Provide data while init MultiModalPredictor if this is not expected."
                )
                self.model.CLASSES = self.checkpoint["meta"]["CLASSES"]
            else:
                warnings.warn(
                    f"CLASSES is not provided and this may cause error. Please provide sample_data_path in predictor's initialization."
                )

        self.name_to_id = self.get_layer_ids()
        self.head_layer_names = [n for n, layer_id in self.name_to_id.items() if layer_id <= 0]

    def set_data_preprocessor_device(self):
        if not self.device:
            self.device = next(self.model.parameters()).device
        if self.device != self.data_preprocessor.device:
            self.data_preprocessor.to(self.device)

    def save(self, save_path: str = "./", tokenizers: Optional[dict] = None):
        weights_save_path = os.path.join(save_path, "model.pth")
        configs_save_path = os.path.join(save_path, "config.py")

        self._save_weights(save_path=weights_save_path)
        self._save_configs(save_path=configs_save_path)

        return save_path

    def _save_weights(self, save_path=None):
        if not save_path:
            save_path = f"./{self.checkpoint_name}_autogluon.pth"

        torch.save({"state_dict": self.model.state_dict(), "meta": {"CLASSES": self.model.CLASSES}}, save_path)  # nosec B614

    def _save_configs(self, save_path=None):
        if not save_path:
            save_path = f"./{self.checkpoint_name}_autogluon.py"

        self.config.dump(save_path)

    def _get_checkpoint_and_config_file(self, checkpoint_name: str = None, config_file: str = None):
        from mim.commands import download as mimdownload

        from ..utils import download, get_pretrain_configs_dir

        logger.debug(f"initializing {checkpoint_name}")

        if not checkpoint_name:
            checkpoint_name = self.checkpoint_name
        if not config_file:
            config_file = self.config_file

        mmdet_configs_dir = get_pretrain_configs_dir(subfolder="detection")

        AG_CUSTOM_MODELS = {
            "faster_rcnn_r50_fpn_1x_voc0712": {
                "url": "https://automl-mm-bench.s3.amazonaws.com/voc_script/faster_rcnn_r50_fpn_1x_voc0712_20220320_192712-54bef0f3.pth",
                "config_file": os.path.join(mmdet_configs_dir, "voc", "faster_rcnn_r50_fpn_1x_voc0712.py"),
            },
            "yolox_nano": {
                "url": "https://github.com/Megvii-BaseDetection/YOLOX/releases/download/0.1.1rc0/yolox_nano.pth",
                "config_file": os.path.join(mmdet_configs_dir, "yolox", "yolox_nano_8xb8-300e_coco.py"),
                "source": "MegVii",
            },
            "yolox_tiny": {
                "url": "https://download.openmmlab.com/mmdetection/v2.0/yolox/yolox_tiny_8x8_300e_coco/yolox_tiny_8x8_300e_coco_20211124_171234-b4047906.pth",
                "config_file": os.path.join(mmdet_configs_dir, "yolox", "yolox_tiny_8xb8-300e_coco.py"),
            },
            "yolox_s": {
                "url": "https://download.openmmlab.com/mmdetection/v2.0/yolox/yolox_s_8x8_300e_coco/yolox_s_8x8_300e_coco_20211121_095711-4592a793.pth",
                "config_file": os.path.join(mmdet_configs_dir, "yolox", "yolox_s_8xb8-300e_coco.py"),
            },
            "yolox_m": {
                "url": "https://github.com/Megvii-BaseDetection/YOLOX/releases/download/0.1.1rc0/yolox_m.pth",  # Megvii weight, need more verifications
                "config_file": os.path.join(mmdet_configs_dir, "yolox", "yolox_m_8xb8-300e_coco.py"),
                "source": "MegVii",
            },
            "yolox_l": {
                "url": "https://download.openmmlab.com/mmdetection/v2.0/yolox/yolox_l_8x8_300e_coco/yolox_l_8x8_300e_coco_20211126_140236-d3bd2b23.pth",
                "config_file": os.path.join(mmdet_configs_dir, "yolox", "yolox_l_8xb8-300e_coco.py"),
            },
            "yolox_l_objects365": {  # TODO: update with better pretrained weights
                "url": "https://automl-mm-bench.s3.amazonaws.com/object_detection/checkpoints/yolox/yolox_l_objects365_temp.pth",
                "config_file": os.path.join(mmdet_configs_dir, "yolox", "yolox_l_8xb8-300e_coco.py"),
            },
            "yolox_x": {
                "url": "https://download.openmmlab.com/mmdetection/v2.0/yolox/yolox_x_8x8_300e_coco/yolox_x_8x8_300e_coco_20211126_140254-1ef88d67.pth",
                "config_file": os.path.join(mmdet_configs_dir, "yolox", "yolox_x_8xb8-300e_coco.py"),
            },
            "dino_swinl_tta": {
                "url": "https://github.com/RistoranteRist/mmlab-weights/releases/download/dino-swinl/dino-5scale_swin-l_8xb2-36e_coco-5486e051.pth",
                "config_file": os.path.join(mmdet_configs_dir, "dino", "dino_swinl_tta.py"),
            },
        }

        if os.path.isfile(checkpoint_name):
            checkpoint_file = checkpoint_name
        elif os.path.isdir(checkpoint_name):
            checkpoint_file = os.path.join(checkpoint_name, "model.pth")
            config_file = os.path.join(checkpoint_name, "config.py")
        else:
            if checkpoint_name in AG_CUSTOM_MODELS:
                # TODO: add sha1_hash
                checkpoint_file = download(
                    url=AG_CUSTOM_MODELS[checkpoint_name]["url"],
                )
                if (
                    "source" in AG_CUSTOM_MODELS[checkpoint_name]
                    and AG_CUSTOM_MODELS[checkpoint_name]["source"] == "MegVii"
                ):
                    checkpoint_file = self.convert_megvii_yolox(checkpoint_file)
            else:
                # download config and checkpoint files using openmim
                checkpoint_file = mimdownload(package="mmdet", configs=[checkpoint_name], dest_root=".")[0]

        if config_file:
            if not os.path.isfile(config_file):
                raise ValueError(f"Invalid checkpoint_name ({checkpoint_name}) or config_file ({config_file}): ")
        else:
            if checkpoint_name in AG_CUSTOM_MODELS:
                config_file = AG_CUSTOM_MODELS[checkpoint_name]["config_file"]
            else:
                try:
                    # download config and checkpoint files using openmim
                    mimdownload(package="mmdet", configs=[checkpoint_name], dest_root=".")
                    config_file = checkpoint_name + ".py"
                except Exception as e:
                    raise ValueError(f"Invalid checkpoint_name ({checkpoint_name}) or config_file ({config_file}): ")

        self.checkpoint_name = checkpoint_name
        self.checkpoint_file = checkpoint_file
        self.config_file = config_file

    def _load_config(self):
        # read config files
        if isinstance(self.config_file, str):
            self.config = mmengine.Config.fromfile(self.config_file)
        else:
            if not isinstance(self.config_file, dict):
                raise ValueError(
                    f"The variable config_file has type {type(self.config_file)}."
                    f"Detection Model's config_file should either be a str of file path, or a dict as config."
                )

    @property
    def image_key(self):
        return f"{self.prefix}_{IMAGE}"

    @property
    def image_valid_num_key(self):
        return f"{self.prefix}_{IMAGE_VALID_NUM}"

    @property
    def label_key(self):
        return f"{self.prefix}_{LABEL}"

    @property
    def image_column_prefix(self):
        return f"{self.image_key}_{COLUMN}"

    @property
    def image_feature_dim(self):
        return self.model.num_features

    def forward(
        self,
        batch,
        mode,
    ):
        """
        Parameters
        ----------
        batch
            A dictionary containing the input mini-batch data.
            We need to use the keys with the model prefix to index required data.
        mode
            "loss" or "predict". TODO: support "tensor"
            https://github.com/open-mmlab/mmdetection/blob/main/mmdet/models/detectors/base.py#L58C1

        Returns
        -------
            A dictionary with bounding boxes.
        """

        self.set_data_preprocessor_device()
        data = self.data_preprocessor(batch)
        rets = self.model(
            inputs=data["inputs"],
            data_samples=data["data_samples"],
            mode=mode,
        )

        if mode == "loss":
            return rets
        elif mode == "predict":
            # for detailed data structure, see https://github.com/open-mmlab/mmdetection/blob/main/mmdet/structures/det_data_sample.py
            return [{BBOX: ret.pred_instances, LABEL: ret.gt_instances} for ret in rets]
        else:
            raise ValueError(f"{mode} mode is not supported.")

    def _parse_losses(self, losses):
        return self.model._parse_losses(losses)

    def _assign_backbone_layers(self):
        """
        Backbone layers are only assigned when we use low lr for backbone and high lr for others
        TODO: Add hyperparameter controlling this if later find any models requiring this setting other than dino.
        """
        if "dino" in self.checkpoint_name.lower():
            self.backbone_layers = ["backbone"]
        else:
            self.backbone_layers = None

    def get_layer_ids(
        self,
    ):
        """
        Assign an id to each layer. Layer ids will be used in layer-wise lr decay.
        Basically, id gradually increases when going from the output end to
        the input end. The layers defined in this class, e.g., head, have id 0.

        Returns
        -------
        A dictionary mapping the layer names (keys) to their ids (values).
        """

        # two stage lr: backbone v.s. else, or head v.s. else
        if self.backbone_layers:
            return self.get_layer_ids_by_backbone()
        else:
            return self.get_layer_ids_by_head()

    def get_layer_ids_by_backbone(
        self,
    ):
        """
        Assign an id to each layer. Layer ids will be used in layer-wise lr decay.
        Basically, id gradually increases when going from the output end to
        the input end. The layers defined in this class, e.g., head, have id 0.

        Currently only head to 0 others to 1.

        Returns
        -------
        A dictionary mapping the layer names (keys) to their ids (values).
        """
        name_to_id = {}
        backbones = self.backbone_layers

        for n, _ in self.named_parameters():
            name_to_id[n] = 0
            for backbone in backbones:
                if backbone in n:
                    name_to_id[n] = 1

        return name_to_id

    def get_layer_ids_by_head(
        self,
    ):
        """
        Assign an id to each layer. Layer ids will be used in layer-wise lr decay.
        Basically, id gradually increases when going from the output end to
        the input end. The layers defined in this class, e.g., head, have id 0.

        Currently only head to 0 others to 1.

        Returns
        -------
        A dictionary mapping the layer names (keys) to their ids (values).
        """
        name_to_id = {}
        # for some models, use head lr in "head" of bbox_head
        # now support: yolov3, faster_rcnn, deformable_detr, yolox, vfnet, centernet, cascade_rcnn, detr, htc, atss, ssd
        registered_head_layers_patterns = [
            "bbox_head.fc_cls",
            "bbox_head.fc_reg",
            "bbox_head.convs_pred",
            "bbox_head.cls_branches",
            "bbox_head.multi_level_conv_cls",
            "bbox_head.multi_level_conv_reg",
            "bbox_head.multi_level_conv_obj",
            "bbox_head.vfnet_cls",
            "bbox_head.heatmap_head",
            "bbox_head.atss_cls",
            "bbox_head.cls_convs",
        ]
        # for other models, use head lr in whole bbox_head
        default_head_layers_patterns = ["bbox_head"]

        head_registered = False
        is_yolox = False
        for n, _ in self.named_parameters():
            name_to_id[n] = 1
            for pattern in registered_head_layers_patterns:
                if pattern in n:
                    name_to_id[n] = 0
                    head_registered = True
                if "bbox_head.multi_level_conv_cls" in n:
                    is_yolox = True

        if not head_registered:
            for n, _ in self.named_parameters():
                name_to_id[n] = 1
                for pattern in default_head_layers_patterns:
                    if pattern in n:
                        name_to_id[n] = 0

        if is_yolox and "use_layer_id" in self.config:
            name_to_id = self.get_yolox_layer_ids()

        return name_to_id

    def get_yolox_layer_ids(self):
        # logic not straight forward, need to print out the model to understand
        name_to_value = {}
        for name, _ in self.named_parameters():
            n = name
            n = n.replace("backbone", "0")
            n = n.replace("neck", "1")
            n = n.replace("bbox_head", "2")

            # backbone
            n = n.replace("stem", "0")

            # neck
            n = n.replace("reduce_layers", "0")
            n = n.replace("top_down_blocks", "1")
            n = n.replace("downsamples", "2")
            n = n.replace("bottom_up_blocks", "3")
            n = n.replace("out_convs", "4")

            n = n.replace("main_conv", "0")
            n = n.replace("short_conv", "1")
            n = n.replace("final_conv", "2")
            n = n.replace("blocks", "3")

            # bbox_head
            n = n.replace("multi_level_cls_convs", "0")
            n = n.replace("multi_level_reg_convs", "0")
            n = n.replace("multi_level_conv_cls", "1")
            n = n.replace("multi_level_conv_reg", "1")
            n = n.replace("multi_level_conv_obj", "1")

            value = int("".join(c for c in n if c.isdigit()).ljust(8, "0"))
            name_to_value[name] = value

        values = list(set(name_to_value.values()))
        values.sort(reverse=True)
        value_to_id = dict(zip(values, range(len(values))))

        name_to_id = {}
        for n, _ in self.named_parameters():
            name_to_id[n] = value_to_id[name_to_value[n]]
        return name_to_id

    def convert_megvii_yolox(self, source_path):
        """
        Convert YOLOX in megvii naming to mmdetection naming.
        Using code script from: https://github.com/haiyang-tju/dl_tools/blob/master/megvii_nano_2_mmdet.py
        """
        sd = source_path

        model_dict = torch.load(sd, map_location=torch.device("cpu"))  # nosec B614
        if "state_dict" in model_dict:
            model_dict = model_dict["state_dict"]
        if "model" in model_dict:
            model_dict = model_dict["model"]

        new_dict = dict()
        for k, v in model_dict.items():
            new_k = k

            if "backbone.backbone." in k:
                new_k = k.replace("backbone.backbone.", "backbone.")
            if "backbone.dark2." in new_k:
                new_k = new_k.replace("backbone.dark2.", "backbone.stage1.")
            if "backbone.dark3." in new_k:
                new_k = new_k.replace("backbone.dark3.", "backbone.stage2.")
            if "backbone.dark4." in new_k:
                new_k = new_k.replace("backbone.dark4.", "backbone.stage3.")
            if "backbone.dark5." in new_k:
                new_k = new_k.replace("backbone.dark5.", "backbone.stage4.")
            if "dconv." in new_k:
                new_k = new_k.replace("dconv.", "depthwise_conv.")
            if "pconv." in new_k:
                new_k = new_k.replace("pconv.", "pointwise_conv.")
            if "backbone.stage1.1.conv1." in new_k:
                new_k = new_k.replace("backbone.stage1.1.conv1.", "backbone.stage1.1.main_conv.")
            if "backbone.stage1.1.conv2." in new_k:
                new_k = new_k.replace("backbone.stage1.1.conv2.", "backbone.stage1.1.short_conv.")
            if "backbone.stage1.1.conv3." in new_k:
                new_k = new_k.replace("backbone.stage1.1.conv3.", "backbone.stage1.1.final_conv.")
            if ".m." in new_k:
                new_k = new_k.replace(".m.", ".blocks.")
            if "backbone.stage2.1.conv1." in new_k:
                new_k = new_k.replace("backbone.stage2.1.conv1.", "backbone.stage2.1.main_conv.")
            if "backbone.stage2.1.conv2." in new_k:
                new_k = new_k.replace("backbone.stage2.1.conv2.", "backbone.stage2.1.short_conv.")
            if "backbone.stage2.1.conv3." in new_k:
                new_k = new_k.replace("backbone.stage2.1.conv3.", "backbone.stage2.1.final_conv.")
            if "backbone.stage3.1.conv1." in new_k:
                new_k = new_k.replace("backbone.stage3.1.conv1.", "backbone.stage3.1.main_conv.")
            if "backbone.stage3.1.conv2." in new_k:
                new_k = new_k.replace("backbone.stage3.1.conv2.", "backbone.stage3.1.short_conv.")
            if "backbone.stage3.1.conv3." in new_k:
                new_k = new_k.replace("backbone.stage3.1.conv3.", "backbone.stage3.1.final_conv.")
            if "backbone.stage4.2.conv1." in new_k:
                new_k = new_k.replace("backbone.stage4.2.conv1.", "backbone.stage4.2.main_conv.")
            if "backbone.stage4.2.conv2." in new_k:
                new_k = new_k.replace("backbone.stage4.2.conv2.", "backbone.stage4.2.short_conv.")
            if "backbone.stage4.2.conv3." in new_k:
                new_k = new_k.replace("backbone.stage4.2.conv3.", "backbone.stage4.2.final_conv.")
            if "backbone.lateral_conv0." in new_k:
                new_k = new_k.replace("backbone.lateral_conv0.", "neck.reduce_layers.0.")
            if "backbone.reduce_conv1." in new_k:
                new_k = new_k.replace("backbone.reduce_conv1.", "neck.reduce_layers.1.")
            if "backbone.C3_p4." in new_k:
                new_k = new_k.replace("backbone.C3_p4.", "neck.top_down_blocks.0.")
            if "neck.top_down_blocks.0.conv1." in new_k:
                new_k = new_k.replace("neck.top_down_blocks.0.conv1.", "neck.top_down_blocks.0.main_conv.")
            if "neck.top_down_blocks.0.conv2." in new_k:
                new_k = new_k.replace("neck.top_down_blocks.0.conv2.", "neck.top_down_blocks.0.short_conv.")
            if "neck.top_down_blocks.0.conv3." in new_k:
                new_k = new_k.replace("neck.top_down_blocks.0.conv3.", "neck.top_down_blocks.0.final_conv.")
            if "backbone.C3_p3." in new_k:
                new_k = new_k.replace("backbone.C3_p3.", "neck.top_down_blocks.1.")
            if "neck.top_down_blocks.1.conv1." in new_k:
                new_k = new_k.replace("neck.top_down_blocks.1.conv1.", "neck.top_down_blocks.1.main_conv.")
            if "neck.top_down_blocks.1.conv2." in new_k:
                new_k = new_k.replace("neck.top_down_blocks.1.conv2.", "neck.top_down_blocks.1.short_conv.")
            if "neck.top_down_blocks.1.conv3." in new_k:
                new_k = new_k.replace("neck.top_down_blocks.1.conv3.", "neck.top_down_blocks.1.final_conv.")

            if "backbone.bu_conv2." in new_k:
                new_k = new_k.replace("backbone.bu_conv2.", "neck.downsamples.0.")
            if "backbone.bu_conv1." in new_k:
                new_k = new_k.replace("backbone.bu_conv1.", "neck.downsamples.1.")

            if "backbone.C3_n3." in new_k:
                new_k = new_k.replace("backbone.C3_n3.", "neck.bottom_up_blocks.0.")
            if "neck.bottom_up_blocks.0.conv1." in new_k:
                new_k = new_k.replace("neck.bottom_up_blocks.0.conv1.", "neck.bottom_up_blocks.0.main_conv.")
            if "neck.bottom_up_blocks.0.conv2." in new_k:
                new_k = new_k.replace("neck.bottom_up_blocks.0.conv2.", "neck.bottom_up_blocks.0.short_conv.")
            if "neck.bottom_up_blocks.0.conv3." in new_k:
                new_k = new_k.replace("neck.bottom_up_blocks.0.conv3.", "neck.bottom_up_blocks.0.final_conv.")
            if "backbone.C3_n4." in new_k:
                new_k = new_k.replace("backbone.C3_n4.", "neck.bottom_up_blocks.1.")
            if "neck.bottom_up_blocks.1.conv1." in new_k:
                new_k = new_k.replace("neck.bottom_up_blocks.1.conv1.", "neck.bottom_up_blocks.1.main_conv.")
            if "neck.bottom_up_blocks.1.conv2." in new_k:
                new_k = new_k.replace("neck.bottom_up_blocks.1.conv2.", "neck.bottom_up_blocks.1.short_conv.")
            if "neck.bottom_up_blocks.1.conv3." in new_k:
                new_k = new_k.replace("neck.bottom_up_blocks.1.conv3.", "neck.bottom_up_blocks.1.final_conv.")

            if "head.stems." in new_k:
                new_k = new_k.replace("head.stems.", "neck.out_convs.")
            if "head.cls_convs." in new_k:
                new_k = new_k.replace("head.cls_convs.", "bbox_head.multi_level_cls_convs.")
            if "head.reg_convs." in new_k:
                new_k = new_k.replace("head.reg_convs.", "bbox_head.multi_level_reg_convs.")
            if "head.cls_preds." in new_k:
                new_k = new_k.replace("head.cls_preds.", "bbox_head.multi_level_conv_cls.")
            if "head.reg_preds." in new_k:
                new_k = new_k.replace("head.reg_preds.", "bbox_head.multi_level_conv_reg.")
            if "head.obj_preds." in new_k:
                new_k = new_k.replace("head.obj_preds.", "bbox_head.multi_level_conv_obj.")

            if "bbox_head.multi_level_conv_cls." in new_k:
                if self.classes:
                    new_dict[new_k] = v[: len(self.classes), ...]  # there take the num_classes
                else:
                    new_dict[new_k] = v
            else:
                new_dict[new_k] = v

        data = {"state_dict": new_dict}

        target_directory = os.path.splitext(sd)[0] + f"_cvt.pth"
        torch.save(data, target_directory)  # nosec B614

        return target_directory



import ast
import copy
import logging
import random
import warnings
from io import BytesIO
from typing import Callable, Dict, List, Optional, Union

import numpy as np
import PIL
import torch
from omegaconf import ListConfig
from PIL import ImageFile
from torch import nn
from torchvision import transforms

from .randaug import RandAugment
from .trivial_augmenter import TrivialAugment

try:
    from torchvision.transforms import InterpolationMode

    BICUBIC = InterpolationMode.BICUBIC
    NEAREST = InterpolationMode.NEAREST
except ImportError:
    BICUBIC = PIL.Image.BICUBIC
    NEAREST = PIL.Image.NEAREST

from ..constants import COLUMN, IMAGE, IMAGE_BASE64_STR, IMAGE_BYTEARRAY, IMAGE_VALID_NUM
from .collator import PadCollator, StackCollator

logger = logging.getLogger(__name__)
ImageFile.LOAD_TRUNCATED_IMAGES = True


class ImageProcessor:
    """
    Prepare image data for the model specified by "prefix". For multiple models requiring image data,
    we need to create a ImageProcessor for each related model so that they will have independent input.
    """

    def __init__(
        self,
        model: nn.Module,
        train_transforms: Union[List[str], Callable, List[Callable]],
        val_transforms: Union[List[str], Callable, List[Callable]],
        max_image_num_per_column: Optional[int] = 1,
        missing_value_strategy: Optional[str] = "zero",
        requires_column_info: Optional[bool] = False,
        dropout: Optional[float] = 0,
    ):
        """
        Parameters
        ----------
        model
            The model for which this processor would be created.
        train_transforms
            A list of image transforms used in training. Note that the transform order matters.
        val_transforms
            A list of image transforms used in validation/test/prediction. Note that the transform order matters.
        max_image_num_per_column
            The maximum number of images one sample can have.
        missing_value_strategy
            How to deal with a missing image. We now support:
            - skip
                Skip this sample
            -zero
                Use an image with zero pixels.
        requires_column_info
            Whether to require feature column information in dataloader.
        """
        logger.debug(f"initializing image processor for model {model.prefix}")
        self.train_transforms = train_transforms
        self.val_transforms = val_transforms
        logger.debug(f"image training transforms: {self.train_transforms}")
        logger.debug(f"image validation transforms: {self.val_transforms}")

        self.prefix = model.prefix
        self.missing_value_strategy = missing_value_strategy
        self.requires_column_info = requires_column_info
        assert 0 <= dropout <= 1
        if dropout > 0:
            logger.debug(f"image dropout probability: {dropout}")
        self.dropout = dropout
        self.size = model.image_size
        self.mean = model.image_mean
        self.std = model.image_std

        self.normalization = transforms.Normalize(self.mean, self.std)
        if max_image_num_per_column <= 0:
            logger.debug(f"max_image_num_per_column {max_image_num_per_column} is reset to 1")
            max_image_num_per_column = 1
        self.max_image_num_per_column = max_image_num_per_column
        logger.debug(f"max_image_num_per_column: {max_image_num_per_column}")

        self.train_processor = self.construct_image_processor(
            image_transforms=self.train_transforms, size=self.size, normalization=self.normalization
        )
        self.val_processor = self.construct_image_processor(
            image_transforms=self.val_transforms, size=self.size, normalization=self.normalization
        )

    @property
    def image_key(self):
        return f"{self.prefix}_{IMAGE}"

    @property
    def image_valid_num_key(self):
        return f"{self.prefix}_{IMAGE_VALID_NUM}"

    @property
    def image_column_prefix(self):
        return f"{self.image_key}_{COLUMN}"

    def collate_fn(self, image_column_names: Optional[List] = None, per_gpu_batch_size: Optional[int] = None) -> Dict:
        """
        Collate images into a batch. Here it pads images since the image number may
        vary from sample to sample. Samples with fewer images will be padded zeros.
        The valid image numbers of samples will be stacked into a vector.
        This function will be used when creating Pytorch DataLoader.

        Returns
        -------
        A dictionary containing one model's collator function for image data.
        """
        fn = {}
        if self.requires_column_info:
            assert image_column_names, "Empty image column names."
            for col_name in image_column_names:
                fn[f"{self.image_column_prefix}_{col_name}"] = StackCollator()

        fn.update(
            {
                self.image_key: PadCollator(pad_val=0),
                self.image_valid_num_key: StackCollator(),
            }
        )

        return fn

    def process_one_sample(
        self,
        images: Dict[str, Union[List[str], List[bytearray]]],
        sub_dtypes: Dict[str, str],
        is_training: bool,
        image_mode: Optional[str] = "RGB",
    ) -> Dict:
        """
        Read images, process them, and stack them. One sample can have multiple images,
        resulting in a tensor of (n, 3, size, size), where n <= max_image_num_per_column is the available image number.

        Parameters
        ----------
        images
            One sample may have multiple image columns in a pd.DataFrame and multiple images
            inside each image column.
        sub_dtypes
            What modality each column belongs to.
        is_training
            Whether to process images in the training mode.
        image_mode
            A string which defines the type and depth of a pixel in the image.
            For example, RGB, RGBA, CMYK, and etc.

        Returns
        -------
        A dictionary containing one sample's images and their number.
        """
        valid_images = []
        zero_images = []
        ret = {}
        column_start = 0

        for per_col_name, per_col_image_raw in images.items():
            for img_raw in per_col_image_raw[: self.max_image_num_per_column]:
                if is_training and self.dropout > 0 and random.uniform(0, 1) <= self.dropout:
                    img = PIL.Image.new(image_mode, (self.size, self.size), color=0)
                    is_zero_img = True
                else:
                    with warnings.catch_warnings():
                        warnings.filterwarnings(
                            "ignore",
                            message=(
                                "Palette images with Transparency expressed in bytes should be converted to RGBA images"
                            ),
                        )
                        is_zero_img = False
                        try:
                            if sub_dtypes.get(per_col_name) in [IMAGE_BYTEARRAY, IMAGE_BASE64_STR]:
                                img_raw = BytesIO(img_raw)

                            with PIL.Image.open(img_raw) as img:
                                img = img.convert(image_mode)
                        except Exception as e:
                            if self.missing_value_strategy.lower() == "zero":
                                logger.debug(f"Using a zero image due to '{e}'")
                                img = PIL.Image.new(image_mode, (self.size, self.size), color=0)
                                is_zero_img = True
                            else:
                                raise e
                if is_training:
                    img = self.train_processor(img)
                else:
                    img = self.val_processor(img)

                if is_zero_img:
                    zero_images.append(img)
                else:
                    valid_images.append(img)

            if self.requires_column_info:
                # only count the valid images since they are put ahead of the zero images in the below returning
                ret[f"{self.image_column_prefix}_{per_col_name}"] = np.array(
                    [column_start, len(valid_images)], dtype=np.int64
                )
                column_start = len(valid_images)

        ret.update(
            {
                self.image_key: torch.tensor([])
                if len(valid_images + zero_images) == 0
                else torch.stack(valid_images + zero_images, dim=0),
                self.image_valid_num_key: len(valid_images),
            }
        )
        return ret

    @staticmethod
    def get_image_transform_funcs(transform_types: Union[List[str], ListConfig, List[Callable]], size: int):
        """
        Parse a list of transform strings into callable objects.

        Parameters
        ----------
        transform_types
            A list of transforms, which can be strings or callable objects.
        size
            Image size.

        Returns
        -------
        A list of transform objects.
        """
        image_transforms = []

        if not transform_types:
            return image_transforms

        if isinstance(transform_types, ListConfig):
            transform_types = list(transform_types)
        elif not isinstance(transform_types, list):
            transform_types = [transform_types]

        if all([isinstance(trans_type, str) for trans_type in transform_types]):
            pass
        elif all([isinstance(trans_type, Callable) for trans_type in transform_types]):
            return copy.copy(transform_types)
        else:
            raise ValueError(
                f"transform_types {transform_types} contain neither all strings nor all callable objects."
            )

        for trans_type in transform_types:
            args = None
            kargs = None
            if "(" in trans_type:
                trans_mode = trans_type[0 : trans_type.find("(")]
                if "{" in trans_type:
                    kargs = ast.literal_eval(trans_type[trans_type.find("{") : trans_type.rfind(")")])
                else:
                    args = ast.literal_eval(trans_type[trans_type.find("(") :])
            else:
                trans_mode = trans_type

            if trans_mode == "resize_to_square":
                image_transforms.append(transforms.Resize((size, size), interpolation=BICUBIC))
            elif trans_mode == "resize_gt_to_square":
                image_transforms.append(transforms.Resize((size, size), interpolation=NEAREST))
            elif trans_mode == "resize_shorter_side":
                image_transforms.append(transforms.Resize(size, interpolation=BICUBIC))
            elif trans_mode == "center_crop":
                image_transforms.append(transforms.CenterCrop(size))
            elif trans_mode == "random_resize_crop":
                image_transforms.append(transforms.RandomResizedCrop(size))
            elif trans_mode == "random_horizontal_flip":
                image_transforms.append(transforms.RandomHorizontalFlip())
            elif trans_mode == "random_vertical_flip":
                image_transforms.append(transforms.RandomVerticalFlip())
            elif trans_mode == "color_jitter":
                if kargs is not None:
                    image_transforms.append(transforms.ColorJitter(**kargs))
                elif args is not None:
                    image_transforms.append(transforms.ColorJitter(*args))
                else:
                    image_transforms.append(transforms.ColorJitter(brightness=0.1, contrast=0.1, saturation=0.1))
            elif trans_mode == "affine":
                if kargs is not None:
                    image_transforms.append(transforms.RandomAffine(**kargs))
                elif args is not None:
                    image_transforms.append(transforms.RandomAffine(*args))
                else:
                    image_transforms.append(
                        transforms.RandomAffine(degrees=15, translate=(0.1, 0.1), scale=(0.9, 1.1))
                    )
            elif trans_mode == "randaug":
                if kargs is not None:
                    image_transforms.append(RandAugment(**kargs))
                elif args is not None:
                    image_transforms.append(RandAugment(*args))
                else:
                    image_transforms.append(RandAugment(2, 9))
            elif trans_mode == "trivial_augment":
                image_transforms.append(TrivialAugment(IMAGE, 30))
            else:
                raise ValueError(f"unknown transform type: {trans_mode}")

        return image_transforms

    def construct_image_processor(
        self,
        image_transforms: Union[List[Callable], List[str]],
        size: int,
        normalization,
    ) -> transforms.Compose:
        """
        Build up an image processor from the provided list of transform types.

        Parameters
        ----------
        image_transforms
            A list of image transform types.
        size
            Image size.
        normalization
            A transforms.Normalize object. When the image is ground truth image, 'normalization=None' should be specified.

        Returns
        -------
        A transforms.Compose object.
        """
        image_transforms = self.get_image_transform_funcs(transform_types=image_transforms, size=size)
        if not any([isinstance(trans, transforms.ToTensor) for trans in image_transforms]):
            image_transforms.append(transforms.ToTensor())
        if (
            not any([isinstance(trans, transforms.Normalize) for trans in image_transforms])
            and normalization is not None
        ):
            image_transforms.append(normalization)
        return transforms.Compose(image_transforms)

    def __call__(
        self,
        images: Dict[str, List[str]],
        sub_dtypes: Dict[str, str],
        is_training: bool,
    ) -> Dict:
        """
        Obtain one sample's images and customized them for a specific model.

        Parameters
        ----------
        images
            Images of one sample.
        sub_dtypes
            The sub data types of all image columns.
        is_training
            Whether to process images in the training mode.

        Returns
        -------
        A dictionary containing one sample's processed images and their number.
        """
        images = {k: [v] if isinstance(v, str) else v for k, v in images.items()}

        return self.process_one_sample(images=images, sub_dtypes=sub_dtypes, is_training=is_training)

    def __getstate__(self):
        odict = self.__dict__.copy()  # get attribute dictionary
        del odict["train_processor"]  # remove augmenter to support pickle
        return odict

    def __setstate__(self, state):
        self.__dict__ = state
        self.train_processor = self.construct_image_processor(
            image_transforms=self.train_transforms,
            size=self.size,
            normalization=self.normalization,
        )



#!/usr/bin/env python
###########################
# This code block is a HACK (!), but is necessary to avoid code duplication. Do NOT alter these lines.
import importlib.util
import os
import platform

from setuptools import setup

filepath = os.path.abspath(os.path.dirname(__file__))
filepath_import = os.path.join(filepath, "..", "core", "src", "autogluon", "core", "_setup_utils.py")
spec = importlib.util.spec_from_file_location("ag_min_dependencies", filepath_import)
ag = importlib.util.module_from_spec(spec)
# Identical to `from autogluon.core import _setup_utils as ag`, but works without `autogluon.core` being installed.
spec.loader.exec_module(ag)
###########################

version = ag.load_version_file()
version = ag.update_version(version)

submodule = "multimodal"
install_requires = [
    # version ranges added in ag.get_dependency_version_ranges()
    "numpy",  # version range defined in `core/_setup_utils.py`
    "scipy",  # version range defined in `core/_setup_utils.py`
    "pandas",  # version range defined in `core/_setup_utils.py`
    "scikit-learn",  # version range defined in `core/_setup_utils.py`
    "Pillow",  # version range defined in `core/_setup_utils.py`
    "tqdm",  # version range defined in `core/_setup_utils.py`
    "boto3",  # version range defined in `core/_setup_utils.py`
    "torch",  # version range defined in `core/_setup_utils.py`
    "lightning",  # version range defined in `core/_setup_utils.py`
    "transformers[sentencepiece]",  # version range defined in `core/_setup_utils.py`
    "accelerate",  # version range defined in `core/_setup_utils.py`
    "requests>=2.30,<3",
    "jsonschema>=4.18,<4.24",
    "seqeval>=1.2.2,<1.3.0",
    "evaluate>=0.4.0,<0.5.0",
    "timm>=0.9.5,<1.0.7",
    "torchvision>=0.16.0,<0.22.0",
    "scikit-image>=0.19.1,<0.26.0",
    "text-unidecode>=1.3,<1.4",
    "torchmetrics>=1.2.0,<1.8",
    "omegaconf>=2.1.1,<2.4.0",
    f"autogluon.core[raytune]=={version}",
    f"autogluon.features=={version}",
    f"autogluon.common=={version}",
    "pytorch-metric-learning>=1.3.0,<2.9",
    "nlpaug>=1.1.10,<1.2.0",
    "nltk>=3.4.5,<4.0",
    "openmim>=0.3.7,<0.4.0",
    "defusedxml>=0.7.1,<0.7.2",
    "jinja2>=3.0.3,<3.2",
    "tensorboard>=2.9,<3",
    "pytesseract>=0.3.9,<0.4",
    "nvidia-ml-py3>=7.352.0, <8.0",
    "pdf2image>=1.17.0,<1.19",
]

install_requires = ag.get_dependency_version_ranges(install_requires)

tests_require = [
    "ruff",
    "datasets>=2.16.0,<3.6.0",
    "onnx>=1.13.0,<1.16.2;platform_system=='Windows'",  # cap at 1.16.1 for issue https://github.com/onnx/onnx/issues/6267
    "onnx>=1.13.0,<1.18.0;platform_system!='Windows'",
    "onnxruntime>=1.17.0,<1.22.0",  # install for gpu system due to https://github.com/autogluon/autogluon/issues/3804
    "onnxruntime-gpu>=1.17.0,<1.22.0;platform_system!='Darwin' and platform_machine!='aarch64'",
    "tensorrt>=8.6.0,<10.9.1;platform_system=='Linux' and python_version<'3.11'",
]

extras_require = {"tests": tests_require}


if __name__ == "__main__":
    ag.create_version_file(version=version, submodule=submodule)
    setup_args = ag.default_setup_args(version=version, submodule=submodule)
    setup_args["package_data"]["autogluon.multimodal"] = [
        "configs/data/*.yaml",
        "configs/model/*.yaml",
        "configs/optim/*.yaml",
        "configs/env/*.yaml",
        "configs/distiller/*.yaml",
        "configs/matcher/*.yaml",
    ]
    setup(
        install_requires=install_requires,
        extras_require=extras_require,
        **setup_args,
    )



checkpoint_config = dict(interval=1)
# yapf:disable
log_config = dict(
    interval=50,
    hooks=[
        dict(type="TextLoggerHook"),
        # dict(type='TensorboardLoggerHook')
    ],
)
# yapf:enable
custom_hooks = [dict(type="NumClassCheckHook")]

dist_params = dict(backend="nccl")
log_level = "INFO"
load_from = None
resume_from = None
workflow = [("train", 1)]

# disable opencv multithreading to avoid system being overloaded
opencv_num_threads = 0
# set multi-process start method as `fork` to speed up the training
mp_start_method = "fork"

# Default setting for scaling LR automatically
#   - `enable` means enable scaling LR automatically
#       or not by default.
#   - `base_batch_size` = (8 GPUs) x (2 samples per GPU).
auto_scale_lr = dict(enable=False, base_batch_size=16)



import logging
import random
from typing import Any, Dict, List, Optional, Union

from torch import nn

from ..constants import CATEGORICAL, COLUMN
from .collator import StackCollator, TupleCollator

logger = logging.getLogger(__name__)


class CategoricalProcessor:
    """
    Prepare categorical data for the model specified by "prefix".
    For multiple models requiring categorical data, we need to create a CategoricalProcessor
    for each related model so that they will have independent input.
    """

    def __init__(
        self,
        model: nn.Module,
        requires_column_info: bool = False,
        dropout: Optional[float] = 0,
    ):
        """
        Parameters
        ----------
        model
            The model for which this processor would be created.
        requires_column_info
            Whether to require feature column information in dataloader.
        """
        logger.debug(f"initializing categorical processor for model {model.prefix}")
        self.prefix = model.prefix
        self.requires_column_info = requires_column_info
        self.num_categories = model.num_categories
        self.dropout = dropout
        assert 0 <= self.dropout <= 1
        if self.dropout > 0:
            logger.debug(f"categorical value dropout probability: {self.dropout}")
            fill_values = {k: v - 1 for k, v in self.num_categories.items()}
            logger.debug(f"dropped values will be replaced by {fill_values}")

    @property
    def categorical_key(self):
        return f"{self.prefix}_{CATEGORICAL}"

    @property
    def categorical_column_prefix(self):
        return f"{self.categorical_key}_{COLUMN}"

    def collate_fn(self, categorical_column_names: Optional[List] = None) -> Dict:
        """
        Collate individual samples into a batch. It stacks categorical features of
        each column independently. This function will be used when creating Pytorch DataLoader.

        Returns
        -------
        A dictionary containing one model's collator function for categorical features.
        """
        fn = {}
        if self.requires_column_info:
            assert categorical_column_names, "Empty categorical column names."
            for col_name in categorical_column_names:
                fn[f"{self.categorical_column_prefix}_{col_name}"] = StackCollator()

        fn[self.categorical_key] = TupleCollator([StackCollator() for _ in range(len(categorical_column_names))])

        return fn

    def process_one_sample(
        self,
        categorical_features: Dict[str, int],
        is_training: bool,
    ) -> Dict:
        """
        Process one sample's categorical features. Assume the categorical features
        are the encoded labels from sklearn' LabelEncoder().

        Parameters
        ----------
        categorical_features
            Categorical features of one sample.
        is_training
            Whether to do processing in the training mode.

        Returns
        -------
        A dictionary containing the processed categorical features.
        """
        ret = {}
        if self.requires_column_info:
            # TODO: consider moving this for loop into __init__() since each sample has the same information.
            for i, col_name in enumerate(categorical_features.keys()):
                ret[f"{self.categorical_column_prefix}_{col_name}"] = i

        if is_training and self.dropout > 0:
            categorical_features_copy = dict()
            for k, v in categorical_features.items():
                if random.uniform(0, 1) <= self.dropout:
                    categorical_features_copy[k] = self.num_categories[k] - 1
                else:
                    categorical_features_copy[k] = v
            categorical_features = categorical_features_copy

        # make sure keys are in the same order
        assert list(categorical_features.keys()) == list(self.num_categories.keys())
        ret[self.categorical_key] = list(categorical_features.values())

        return ret

    def __call__(
        self,
        categorical_features: Dict[str, int],
        sub_dtypes: Dict[str, str],
        is_training: bool,
    ) -> Dict:
        """
        Extract one sample's categorical features and customize it for a specific model.

        Parameters
        ----------
        categorical_features
            Categorical features of one sample.
        sub_dtypes
            The sub data types of all categorical columns.
        is_training
            Whether to do processing in the training mode.

        Returns
        -------
        A dictionary containing one sample's processed categorical features.
        """
        return self.process_one_sample(
            categorical_features=categorical_features,
            is_training=is_training,
        )



"""
Convert VOC format dataset to COCO.
Reference: https://github.com/yukkyo/voc2coco/blob/master/voc2coco.py
With changes:
1. id stored as int by default
2. provide only root_dir, and corresponding simplification
3. split train/val/test
4. Use defusedxml.ElementTree for security concern
5. remove invalid images

To use:
    If you'd like to customize train/val/test ratio. Note test_ratio = 1 - train_ratio - val_ratio.
        python3 -m autogluon.multimodal.cli.voc2coco --root_dir <root_dir> --train_ratio <train_ratio> --val_ratio <val_ratio>
    If you'd like to use the dataset provided train/val/test splits:
        python3 -m autogluon.multimodal.cli.voc2coco --root_dir <root_dir>
"""

import argparse
import json
import os
import random
import re
import subprocess
from typing import Dict, List

import defusedxml.ElementTree as ET
from tqdm import tqdm

from autogluon.multimodal.utils.object_detection import dump_voc_classes, dump_voc_xml_files, process_voc_annotations

DEFAULT_EXT = ".jpg"
MIN_AREA = 4


def get_label2id(labels_path: str) -> Dict[str, int]:
    """id is 1 start"""
    with open(labels_path, "r") as f:
        labels_str = f.read().split()
    labels_ids = list(range(1, len(labels_str) + 1))
    return dict(zip(labels_str, labels_ids))


def get_annpaths(
    root_dir: str,
    annpaths_list_path: str = None,
    train_ratio=0.6,
    val_ratio=0.2,
) -> Dict:
    if annpaths_list_path is not None:
        with open(annpaths_list_path, "r") as f:
            ann_paths = f.read().split()
        random.shuffle(ann_paths)
        N = len(ann_paths)
        num_train = int(N * train_ratio)
        num_val = int(N * val_ratio)
        return {
            "usersplit_train": ann_paths[:num_train],
            "usersplit_val": ann_paths[num_train : num_train + num_val],
            "usersplit_test": ann_paths[num_train + num_val :],
        }
    else:
        ann_ids_folder = os.path.join(root_dir, "ImageSets", "Main")
        ann_dir_path = os.path.join(root_dir, "Annotations")
        ann_paths = {}
        for ann_ids_filename in os.listdir(ann_ids_folder):
            ann_ids_path = os.path.join(ann_ids_folder, ann_ids_filename)
            if os.path.isfile(ann_ids_path) and ann_ids_filename[-4:] == ".txt":
                ann_ids_name = ann_ids_filename[:-4]

                with open(ann_ids_path, "r") as f:
                    rows = f.readlines()
                    if not rows:
                        print(f"Skipping {ann_ids_path}: file is empty")
                    else:
                        ann_ids = []
                        for r in rows:
                            data = r.strip().split()
                            if len(data) == 1:  # Each row is an annotation id
                                ann_ids.append(data[0])
                            elif (
                                len(data) == 2
                            ):  # Each row contains an annotation id and a flag (0 if we do not use this annotation in this split, and 1 if we use it)
                                ann_id, used = data
                                if int(used) == 1:
                                    ann_ids.append(ann_id)
                            else:
                                print(
                                    f"Skipping {ann_ids_path}: file format not recognized. Make sure your annotation follows "
                                    f"VOC format!"
                                )
                                break

                        ann_paths[ann_ids_name] = [os.path.join(ann_dir_path, aid + ".xml") for aid in ann_ids]
        return ann_paths


def get_image_info(annotation_root, extract_num_from_imgid=True):
    path = annotation_root.findtext("path")
    if path is None:
        filename = annotation_root.findtext("filename")
    else:
        filename = os.path.basename(path)
    img_name = os.path.basename(filename)
    if not img_name[-4:] in [".jpg", ".png"]:
        img_name = img_name + DEFAULT_EXT
    img_id = os.path.splitext(img_name)[0]
    if extract_num_from_imgid and isinstance(img_id, str):
        img_id = int("".join(re.findall(r"\d+", img_id)))

    size = annotation_root.find("size")
    width = int(size.findtext("width"))
    height = int(size.findtext("height"))

    image_info = {"file_name": os.path.join("JPEGImages", img_name), "height": height, "width": width, "id": img_id}
    return image_info


def get_coco_annotation_from_obj(obj, label2id):
    label = obj.findtext("name")
    assert label in label2id, f"Error: {label} is not in label2id !"
    category_id = label2id[label]
    bndbox = obj.find("bndbox")
    xmin = int(float(bndbox.findtext("xmin")))
    ymin = int(float(bndbox.findtext("ymin")))
    xmax = int(float(bndbox.findtext("xmax")))
    ymax = int(float(bndbox.findtext("ymax")))
    if xmin >= xmax or ymin >= ymax:
        return {}
    o_width = xmax - xmin
    o_height = ymax - ymin
    area = o_width * o_height
    if area <= MIN_AREA:
        return {}
    ann = {
        "area": o_width * o_height,
        "iscrowd": 0,
        "bbox": [xmin, ymin, o_width, o_height],
        "category_id": category_id,
        "ignore": 0,
        "segmentation": [],  # This script is not for segmentation
    }
    return ann


def convert_xmls_to_cocojson(
    annotation_paths: List[str],
    label2id: Dict[str, int],
    output_jsonpath: str,
    extract_num_from_imgid: bool = True,
    root_dir: str = "./",
):
    output_json_dict = {"images": [], "type": "instances", "annotations": [], "categories": []}
    bnd_id = 1  # START_BOUNDING_BOX_ID, TODO input as args ?
    print("Start converting !")
    for a_path in tqdm(annotation_paths):
        # Read annotation xml
        try:
            ann_tree = ET.parse(a_path)
        except:
            ann_tree = ET.parse(os.path.join(root_dir, a_path))
        ann_root = ann_tree.getroot()

        img_info = get_image_info(annotation_root=ann_root, extract_num_from_imgid=extract_num_from_imgid)
        img_id = img_info["id"]

        valid_image = False  # remove image without bounding box to speed up mAP calculation
        for obj in ann_root.findall("object"):
            ann = get_coco_annotation_from_obj(obj=obj, label2id=label2id)
            if ann:
                ann.update({"image_id": img_id, "id": bnd_id})
                output_json_dict["annotations"].append(ann)
                bnd_id = bnd_id + 1
                valid_image = True

        if valid_image:
            output_json_dict["images"].append(img_info)

    for label, label_id in label2id.items():
        category_info = {"supercategory": "none", "id": label_id, "name": label}
        output_json_dict["categories"].append(category_info)

    with open(output_jsonpath, "w") as f:
        output_json = json.dumps(output_json_dict)
        f.write(output_json)
        print(f"The COCO format annotation is saved to {output_jsonpath}")


def main():
    parser = argparse.ArgumentParser(description="This script support converting voc format xmls to coco format json")
    parser.add_argument("--root_dir", type=str, default=None, help="path to VOC format dataset root")
    parser.add_argument("--train_ratio", type=float, default=None, help="training set ratio")
    parser.add_argument("--val_ratio", type=float, default=None, help="validation set ratio")
    parser.add_argument("--ext", type=str, default=".jpg", help="default extension for image file")
    parser.add_argument("--min_area", type=int, default=4, help="min area for a valid bounding box")
    parser.add_argument(
        "--not_extract_num_from_imgid", action="store_true", help="Extract image number from the image filename"
    )
    args = parser.parse_args()

    annpaths_list_path = None
    DEFAULT_EXT = args.ext
    MIN_AREA = args.min_area
    assert DEFAULT_EXT in [".jpg", ".png"]

    if not args.root_dir:
        raise ValueError("Must specify the root of the VOC format dataset.")
    if args.train_ratio is not None:
        assert args.train_ratio >= 0
        assert args.val_ratio >= 0
        assert args.train_ratio + args.val_ratio <= 1
        annpaths_list_path = os.path.join(args.root_dir, "pathlist.txt")
        ## generate pathlist.txt containing all xml file paths
        dump_voc_xml_files(
            voc_annotation_path=os.path.join(args.root_dir, "Annotations"),
            voc_annotation_xml_output_path=annpaths_list_path,
        )

        assert os.path.exists(annpaths_list_path), "FatalError: pathlist.txt does not exist!"

    labels_path = os.path.join(args.root_dir, "labels.txt")
    ## generate labels.txt containing all unique class names
    dump_voc_classes(
        voc_annotation_path=os.path.join(args.root_dir, "Annotations"), voc_class_names_output_path=labels_path
    )

    assert os.path.exists(labels_path), "FatalError: labels.txt does not exist!"

    output_path_fmt = os.path.join(args.root_dir, "Annotations", "%s_cocoformat.json")

    label2id = get_label2id(labels_path=labels_path)
    ann_paths = get_annpaths(
        root_dir=args.root_dir,
        annpaths_list_path=annpaths_list_path,
        train_ratio=args.train_ratio,
        val_ratio=args.val_ratio,
    )
    for mode in ann_paths.keys():
        convert_xmls_to_cocojson(
            annotation_paths=ann_paths[mode],
            label2id=label2id,
            output_jsonpath=output_path_fmt % mode,
            extract_num_from_imgid=(not args.not_extract_num_from_imgid),
            root_dir=args.root_dir,
        )


if __name__ == "__main__":
    main()



import logging
from typing import Any, Dict, List, Optional, Tuple, Union

import torch
from omegaconf import DictConfig, OmegaConf
from pytorch_metric_learning import distances, losses, miners
from torch import nn
from transformers.models.mask2former.modeling_mask2former import Mask2FormerConfig, Mask2FormerLoss

from ...constants import (
    BINARY,
    CONTRASTIVE_LOSS,
    COSINE_SIMILARITY,
    FEW_SHOT_CLASSIFICATION,
    MULTI_NEGATIVES_SOFTMAX_LOSS,
    MULTICLASS,
    NER,
    OBJECT_DETECTION,
    PAIR_MARGIN_MINER,
    REGRESSION,
    SEMANTIC_SEGMENTATION,
)
from .bce_loss import BBCEWithLogitLoss
from .focal_loss import FocalLoss
from .lemda_loss import LemdaLoss
from .softmax_losses import MultiNegativesSoftmaxLoss, SoftTargetCrossEntropy
from .structure_loss import StructureLoss

logger = logging.getLogger(__name__)


def get_loss_func(
    problem_type: str,
    mixup_active: Optional[bool] = None,
    loss_func_name: Optional[str] = None,
    config: Optional[DictConfig] = None,
    **kwargs,
):
    """
    Choose a suitable Pytorch loss module based on the provided problem type.

    Parameters
    ----------
    problem_type
        Type of problem.
    mixup_active
        The activation determining whether to use mixup.
    loss_func_name
        The name of the function the user wants to use.
    config
        The optimization configs containing values such as i.e. optim.loss_func
        An example purpose of this config here is to pass through the parameters for focal loss, i.e.:
            alpha = optim.focal_loss.alpha
    Returns
    -------
    A Pytorch loss module.
    """
    if problem_type in [BINARY, MULTICLASS]:
        if mixup_active:
            loss_func = SoftTargetCrossEntropy()
        else:
            if loss_func_name is not None and loss_func_name.lower() == "focal_loss":
                loss_func = FocalLoss(
                    alpha=config.focal_loss.alpha,
                    gamma=config.focal_loss.gamma,
                    reduction=config.focal_loss.reduction,
                )
            else:
                loss_func = nn.CrossEntropyLoss(label_smoothing=config.label_smoothing)
                logger.debug(f"loss_func.label_smoothing: {loss_func.label_smoothing}")
    elif problem_type == REGRESSION:
        if loss_func_name is not None:
            if "bcewithlogitsloss" in loss_func_name.lower():
                loss_func = nn.BCEWithLogitsLoss()
            else:
                loss_func = nn.MSELoss()
        else:
            loss_func = nn.MSELoss()
    elif problem_type == NER:
        loss_func = nn.CrossEntropyLoss(ignore_index=0)
    elif problem_type in [None, OBJECT_DETECTION, FEW_SHOT_CLASSIFICATION]:
        return None
    elif problem_type == SEMANTIC_SEGMENTATION:
        if "structure_loss" in loss_func_name.lower():
            loss_func = StructureLoss()
        elif "balanced_bce" in loss_func_name.lower():
            loss_func = BBCEWithLogitLoss()
        elif "mask2former_loss" in loss_func_name.lower():
            weight_dict = {
                "loss_cross_entropy": config.mask2former_loss.loss_cross_entropy_weight,
                "loss_mask": config.mask2former_loss.loss_mask_weight,
                "loss_dice": config.mask2former_loss.loss_dice_weight,
            }
            loss_func = Mask2FormerLoss(
                config=Mask2FormerConfig(num_labels=kwargs["num_classes"]), weight_dict=weight_dict
            )
        else:
            loss_func = nn.BCEWithLogitsLoss()
    else:
        raise NotImplementedError

    return loss_func


def get_metric_learning_distance_func(
    name: str,
):
    """
    Return one pytorch metric learning's distance function based on its name.

    Parameters
    ----------
    name
        distance function name

    Returns
    -------
    A distance function from the pytorch metric learning package.
    """
    if name.lower() == COSINE_SIMILARITY:
        return distances.CosineSimilarity()
    else:
        raise ValueError(f"Unknown distance measure: {name}")


def infer_matcher_loss(data_format: str, problem_type: str):
    """
    Infer the loss type to train the matcher.

    Parameters
    ----------
    data_format
        The training data format, e.g., pair or triplet.
    problem_type
        Type of problem.

    Returns
    -------
    The loss name.
    """
    if data_format == "pair":
        if problem_type is None:
            return [MULTI_NEGATIVES_SOFTMAX_LOSS]
        elif problem_type == BINARY:
            return [CONTRASTIVE_LOSS]
        elif problem_type == REGRESSION:
            return ["cosine_similarity_loss"]
        else:
            raise ValueError(f"Unsupported data format {data_format} with problem type {problem_type}")
    elif data_format == "triplet":
        if problem_type is None:
            return [MULTI_NEGATIVES_SOFTMAX_LOSS]
        else:
            raise ValueError(f"Unsupported data format {data_format} with problem type {problem_type}")
    else:
        raise ValueError(f"Unsupported data format: {data_format}")


def get_matcher_loss_func(
    data_format: str,
    problem_type: str,
    loss_type: Optional[str] = None,
    pos_margin: Optional[float] = None,
    neg_margin: Optional[float] = None,
    distance_type: Optional[str] = None,
):
    """
    Return a list of pytorch metric learning's loss functions based on their names.

    Parameters
    ----------
    data_format
        The training data format, e.g., pair or triplet.
    problem_type
        Type of problem.
    loss_type
        The provided loss type.
    pos_margin
        The positive margin in computing the metric learning loss.
    neg_margin
        The negative margin in computing the metric learning loss.
    distance_type
        The distance function type.

    Returns
    -------
    A loss function of metric learning.
    """

    allowable_loss_types = infer_matcher_loss(data_format=data_format, problem_type=problem_type)
    if loss_type is not None:
        assert loss_type in allowable_loss_types, f"data format {data_format} can't use loss {loss_type}."
    else:
        loss_type = allowable_loss_types[0]

    if loss_type.lower() == CONTRASTIVE_LOSS:
        return losses.ContrastiveLoss(
            pos_margin=pos_margin,
            neg_margin=neg_margin,
            distance=get_metric_learning_distance_func(distance_type),
        )
    elif loss_type.lower() == MULTI_NEGATIVES_SOFTMAX_LOSS:
        return MultiNegativesSoftmaxLoss(
            local_loss=True,
            gather_with_grad=True,
            cache_labels=False,
        )
    else:
        raise ValueError(f"Unknown metric learning loss: {loss_type}")


def get_matcher_miner_func(
    miner_type: str,
    pos_margin: float,
    neg_margin: float,
    distance_type: str,
):
    """
    Return a pytorch metric learning's miner functions based on their names.
    The miners are used to mine the positive and negative examples.

    Parameters
    ----------
    miner_type
        The miner function type.
    pos_margin
        The positive margin used by the miner function.
    neg_margin
        The negative margin used by the miner function.
    distance_type
        The distance function type.

    Returns
    -------
    A miner function to mine positive and negative samples.
    """
    if miner_type.lower() == PAIR_MARGIN_MINER:
        return miners.PairMarginMiner(
            pos_margin=pos_margin,
            neg_margin=neg_margin,
            distance=get_metric_learning_distance_func(distance_type),
        )
    else:
        raise ValueError(f"Unknown metric learning miner: {miner_type}")


def generate_metric_learning_labels(
    num_samples: int,
    match_label: int,
    labels: torch.Tensor,
):
    """
    Generate labels to compute the metric learning loss of one mini-batch.
    For n samples, it generates 2*n labels since each match has two sides, each of which
    has one label. If we know the matching label, then it determines the two sides' labels
    according to whether their label is the matching label. If the matching label is None,
    it assigns a unique label for each side.

    Parameters
    ----------
    num_samples
        number of samples.
    match_label
        The matching label, which can be None.
    labels
        The sample labels used in the supervised setting. It's required only when match_label is not None.

    Returns
    -------
    The labels used in computing the metric learning loss.
    """
    device = labels.device
    labels_1 = torch.arange(num_samples, device=device)

    if match_label is not None:
        labels_2 = torch.arange(num_samples, num_samples * 2, device=device)
        # users need to specify the match_label based on the raw label's semantic meaning.
        mask = labels == match_label
        labels_2[mask] = labels_1[mask]
    else:
        labels_2 = torch.arange(num_samples, device=device)

    metric_learning_labels = torch.cat([labels_1, labels_2], dim=0)

    return metric_learning_labels


def get_aug_loss_func(config: Optional[DictConfig] = None, problem_type: Optional[str] = None):
    """
    Return the loss function for lemda augmentation

    Parameters
    ----------
    config
        The optimization configuration.
    problem_type
        Problem type (binary, multimclass, or regression)

    Returns
    -------
    Augmentation loss function.
    """
    loss_func = None
    if config.lemda.turn_on:
        loss_func = LemdaLoss(
            mse_weight=config.lemda.mse_weight,
            kld_weight=config.lemda.kld_weight,
            consist_weight=config.lemda.consist_weight,
            consist_threshold=config.lemda.consist_threshold,
            problem_type=problem_type,
        )

    return loss_func
